# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import math
from typing import List, Optional, Union

import torch
import torch.nn as nn
from scipy.linalg import hadamard

from fastestimator.util.util import to_list


class HadamardCode(nn.Module):
    """A layer for applying an error correcting code to your outputs.

    This class is intentionally not @traceable (models and layers are handled by a different process).

    See 'https://papers.nips.cc/paper/9070-error-correcting-output-codes-improve-probability-estimation-and-adversarial-
    robustness-of-deep-neural-networks'. Note that for best effectiveness, the model leading into this layer should be
    split into multiple independent chunks, whose outputs this layer can combine together in order to perform the code
    lookup.

    ```python
    # Use as a drop-in replacement for your softmax layer:
    def __init__(self, classes):
        self.fc1 = nn.Linear(1024, 64)
        self.fc2 = nn.Linear(64, classes)
    def forward(self, x):
        x = fn.relu(self.fc1(x))
        x = fn.softmax(self.fc2(x), dim=-1)
    #   ----- vs ------
    def __init__(self, classes):
        self.fc1 = nn.Linear(1024, 64)
        self.fc2 = HadamardCode(64, classes)
    def forward(self, x):
        x = fn.relu(self.fc1(x))
        x = self.fc2(x)
    ```

    ```python
    # Use to combine multiple feature heads for a final output (biggest adversarial hardening benefit):
    def __init__(self, classes):
        self.fc1 = nn.ModuleList([nn.Linear(1024, 16) for _ in range(4)])
        self.fc2 = HadamardCode([16]*4, classes)
    def forward(self, x):
        x = [fn.relu(fc(x)) for fc in self.fc1]
        x = self.fc2(x)
    ```

    Args:
        in_features: How many input features there are (inputs should be of shape (Batch, N) or [(Batch, N), ...]).
        n_classes: How many output classes to map onto.
        code_length: How long of an error correcting code to use. Should be a positive multiple of 2. If not provided,
            the smallest power of 2 which is >= `n_outputs` will be used, or 16 if the latter is larger.
        max_prob: The maximum probability that can be assigned to a class. For numeric stability this must be less than
            1.0. Intuitively it makes sense to keep this close to 1, but to get adversarial training benefits it should
            be noticeably less than 1, for example 0.95 or even 0.8.
        power: The power parameter to be used by Inverse Distance Weighting when transforming Hadamard class distances
            into a class probability distribution. A value of 1.0 gives an intuitive mapping to probabilities, but small
            values such as 0.25 appear to give slightly better adversarial benefits. Large values like 2 or 3 give
            slightly faster convergence at the expense of adversarial performance. Must be greater than zero.

    Raises:
        ValueError: If `code_length` is invalid.
    """
    heads: Union[nn.ModuleList, nn.Module]

    def __init__(self,
                 in_features: Union[int, List[int]],
                 n_classes: int,
                 code_length: Optional[int] = None,
                 max_prob: float = 0.95,
                 power: float = 1.0) -> None:
        super().__init__()
        self.n_classes = n_classes
        if code_length is None:
            code_length = max(16, 1 << (n_classes - 1).bit_length())
        if code_length <= 0 or (code_length & (code_length - 1) != 0):
            raise ValueError(f"code_length must be a positive power of 2, but got {code_length}.")
        if code_length < n_classes:
            raise ValueError(f"code_length must be >= n_classes, but got {code_length} and {n_classes}.")
        self.code_length = code_length
        if power <= 0:
            raise ValueError(f"power must be positive, but got {power}.")
        self.power = nn.Parameter(torch.tensor(power), requires_grad=False)
        if not 0.0 < max_prob < 1.0:
            raise ValueError(f"max_prob must be in the range (0, 1), but got {max_prob}")
        self.eps = nn.Parameter(
            torch.tensor(self.code_length * math.pow((1.0 - max_prob) / (max_prob * (self.n_classes - 1)), 1 / power)),
            requires_grad=False)
        labels = hadamard(self.code_length)
        # Cut off 0th column b/c it's constant. It would also be possible to make the column sign alternate, but that
        # would break the symmetry between rows in the code.
        labels = labels[:self.n_classes, 1:]
        self.labels = nn.Parameter(torch.tensor(labels, dtype=torch.float32), requires_grad=False)
        in_features = to_list(in_features)
        if len(in_features) > code_length - 1:
            raise ValueError(f"Too many input heads {len(in_features)} for the given code length {self.code_length}.")
        head_sizes = [self.code_length // len(in_features) for _ in range(len(in_features))]
        head_sizes[0] = head_sizes[0] + self.code_length - sum(head_sizes)
        head_sizes[0] = head_sizes[0] - 1  # We're going to cut off the 0th column from the code
        self.heads = nn.ModuleList([
            nn.Linear(in_features=in_feat, out_features=out_feat) for in_feat, out_feat in zip(in_features, head_sizes)
        ])

    def forward(self, x: List[torch.Tensor]) -> torch.Tensor:
        # can't have forward function call subfunctions otherwise will fail on multi-gpu
        if isinstance(x, list):
            x = [head(tensor) for head, tensor in zip(self.heads, x)]
            x = torch.cat(x, dim=-1)
        else:
            x = self.heads[0](x)
        x = torch.tanh(x)
        # Compute L1 distance
        x = torch.max(torch.sum(torch.abs(torch.unsqueeze(x, dim=1) - self.labels), dim=-1), self.eps)
        # Inverse Distance Weighting
        x = 1.0 / torch.pow(x, self.power)
        x = torch.div(x, torch.sum(x, dim=-1).view(-1, 1))
        return x



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np

from fastestimator.dataset import GeneratorDataset


def inputs():
    while True:
        yield {'x': np.random.rand(4), 'y': np.random.randint(2)}


class TestGeneratorDataset(unittest.TestCase):
    def test_dataset(self):
        dataset = GeneratorDataset(generator=inputs(), samples_per_epoch=10)

        self.assertEqual(len(dataset), 10)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import zipfile
from glob import glob
from pathlib import Path
from typing import Optional

import pandas as pd
import wget

from fastestimator.dataset.csv_dataset import CSVDataset
from fastestimator.util.wget_util import bar_custom, callback_progress

wget.callback_progress = callback_progress


def load_data(root_dir: Optional[str] = None) -> CSVDataset:
    """Load and return the montgomery dataset.

    Sourced from http://openi.nlm.nih.gov/imgs/collections/NLM-MontgomeryCXRSet.zip. This method will download the data
        to local storage if the data has not been previously downloaded.

    Args:
        root_dir: The path to store the downloaded data. When `path` is not provided, the data will be saved into
            `fastestimator_data` under the user's home directory.

    Returns:
        train_data
    """
    home = str(Path.home())

    if root_dir is None:
        root_dir = os.path.join(home, 'fastestimator_data', 'Montgomery')
    else:
        root_dir = os.path.join(os.path.abspath(root_dir), 'Montgomery')
    os.makedirs(root_dir, exist_ok=True)

    csv_path = os.path.join(root_dir, "montgomery.csv")
    data_compressed_path = os.path.join(root_dir, 'NLM-MontgomeryCXRSet.zip')
    extract_folder_path = os.path.join(root_dir, 'MontgomerySet')

    if not os.path.exists(extract_folder_path):
        # download
        if not os.path.exists(data_compressed_path):
            print("Downloading data to {}".format(root_dir))
            wget.download('http://openi.nlm.nih.gov/imgs/collections/NLM-MontgomeryCXRSet.zip',
                          root_dir,
                          bar=bar_custom)

        # extract
        print("\nExtracting file ...")
        with zipfile.ZipFile(data_compressed_path, 'r') as zip_file:
            # There's some garbage data from macOS in the zip file that gets filtered out here
            zip_file.extractall(root_dir, filter(lambda x: x.startswith("MontgomerySet/"), zip_file.namelist()))

    # glob and generate csv
    if not os.path.exists(csv_path):
        img_list = glob(os.path.join(extract_folder_path, 'CXR_png', '*.png'))
        df = pd.DataFrame(data={'image': img_list})
        df['image'] = df['image'].apply(lambda x: os.path.relpath(x, root_dir))
        df['image'] = df['image'].apply(os.path.normpath)
        df['mask_left'] = df['image'].apply(lambda x: x.replace('CXR_png', os.path.join('ManualMask', 'leftMask')))
        df['mask_right'] = df['image'].apply(lambda x: x.replace('CXR_png', os.path.join('ManualMask', 'rightMask')))
        df.to_csv(csv_path, index=False)

    return CSVDataset(csv_path)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import tensorflow as tf

import fastestimator as fe


class TestHadamardCode(unittest.TestCase):
    def test_single_input(self):
        batch_size = 27
        n_features = 123
        n_classes = 10
        x = tf.ones((batch_size, n_features))
        layer = fe.layers.tensorflow.HadamardCode(n_classes=n_classes)
        y = layer(x)
        self.assertEqual(tf.TensorShape(dims=(batch_size, n_classes)), y.shape)
        for row in y:
            # Should have produced a probability vector, so each row must be equal to 1
            self.assertAlmostEqual(1, tf.reduce_sum(row).numpy(), delta=0.000001)

    def test_multi_input(self):
        batch_size = 32
        n_features = 18
        n_classes = 18
        x = [tf.ones((batch_size, n_features)) for _ in range(6)]
        layer = fe.layers.tensorflow.HadamardCode(n_classes=n_classes)
        y = layer(x)
        self.assertEqual(tf.TensorShape(dims=(batch_size, n_classes)), y.shape)
        for row in y:
            # Should have produced a probability vector, so each row must be equal to 1
            self.assertAlmostEqual(1, tf.reduce_sum(row).numpy(), delta=0.000001)

    def test_longer_code(self):
        batch_size = 16
        n_features = 11
        n_classes = 13
        x = tf.ones((batch_size, n_features))
        layer = fe.layers.tensorflow.HadamardCode(n_classes=n_classes, code_length=256)
        y = layer(x)
        self.assertEqual(tf.TensorShape(dims=(batch_size, n_classes)), y.shape)
        for row in y:
            # Should have produced a probability vector, so each row must be equal to 1
            self.assertAlmostEqual(1, tf.reduce_sum(row).numpy(), delta=0.000001)

    def test_shorter_code(self):
        n_classes = 10
        self.assertRaises(ValueError, lambda: fe.layers.tensorflow.HadamardCode(n_classes=n_classes, code_length=8))

    def test_non_power_of_two_code_length(self):
        n_classes = 10
        self.assertRaises(ValueError, lambda: fe.layers.tensorflow.HadamardCode(n_classes=n_classes, code_length=18))

    def test_negative_code_length(self):
        self.assertRaises(ValueError, lambda: fe.layers.tensorflow.HadamardCode(n_classes=5, code_length=-2))



import tempfile
import unittest

import numpy as np

import fastestimator as fe
from fastestimator.dataset import NumpyDataset
from fastestimator.test.unittest_util import sample_system_object, sample_system_object_torch


class TestDataset(NumpyDataset):
    def __init__(self, data, var):
        super().__init__(data)
        self.var = var


class TestBatchDataset(unittest.TestCase):
    def test_save_and_load_state_with_batch_dataset_tf(self):
        def instantiate_system():
            system = sample_system_object()
            x_train = np.ones((2, 28, 28, 3))
            y_train = np.ones((2, ))
            ds = TestDataset(data={'x': x_train, 'y': y_train}, var=1)
            train_data = fe.dataset.BatchDataset(datasets=[ds, ds], num_samples=[1, 1])
            system.pipeline = fe.Pipeline(train_data=train_data, batch_size=2)
            return system

        system = instantiate_system()

        # make some change
        new_var = 2
        system.pipeline.data["train"][None].datasets[0].var = new_var

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_path)

        # reinstantiate system and load the state
        system = instantiate_system()
        system.load_state(save_path)
        loaded_var = system.pipeline.data["train"][None].datasets[0].var

        self.assertEqual(loaded_var, new_var)

    def test_save_and_load_state_with_batch_dataset_torch(self):
        def instantiate_system():
            system = sample_system_object_torch()
            x_train = np.ones((2, 3, 28, 28))
            y_train = np.ones((2, ))
            ds = TestDataset(data={'x': x_train, 'y': y_train}, var=1)
            train_data = fe.dataset.BatchDataset(datasets=[ds, ds], num_samples=[1, 1])
            system.pipeline = fe.Pipeline(train_data=train_data, batch_size=2)
            return system

        system = instantiate_system()

        # make some change
        new_var = 2
        system.pipeline.data["train"][None].datasets[0].var = new_var

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_path)

        # reinstantiate system and load the state
        system = instantiate_system()
        system.load_state(save_path)
        loaded_var = system.pipeline.data["train"][None].datasets[0].var

        self.assertEqual(loaded_var, new_var)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Collection, TypeVar, Union

import numpy as np
import tensorflow as tf
import torch

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor, np.ndarray)


def to_tensor(data: Union[Collection, Tensor, float, int, None], target_type: str) -> Union[Collection, Tensor, None]:
    """Convert tensors within a collection of `data` to a given `target_type` recursively.

    This method can be used with Numpy data:
    ```python
    data = {"x": np.ones((10,15)), "y":[np.ones((4)), np.ones((5, 3))], "z":{"key":np.ones((2,2))}}
    t = fe.backend.to_tensor(data, target_type='tf')
    # {"x": <tf.Tensor>, "y":[<tf.Tensor>, <tf.Tensor>], "z": {"key": <tf.Tensor>}}
    p = fe.backend.to_tensor(data, target_type='torch')
    # {"x": <torch.Tensor>, "y":[<torch.Tensor>, <torch.Tensor>], "z": {"key": <torch.Tensor>}}
    ```

    This method can be used with TensorFlow tensors:
    ```python
    data = {"x": tf.ones((10,15)), "y":[tf.ones((4)), tf.ones((5, 3))], "z":{"key":tf.ones((2,2))}}
    p = fe.backend.to_tensor(data, target_type='torch')
    # {"x": <torch.Tensor>, "y":[<torch.Tensor>, <torch.Tensor>], "z": {"key": <torch.Tensor>}}
    ```

    This method can be used with PyTorch tensors:
    ```python
    data = {"x": torch.ones((10,15)), "y":[torch.ones((4)), torch.ones((5, 3))], "z":{"key":torch.ones((2,2))}}
    t = fe.backend.to_tensor(data, target_type='tf')
    # {"x": <tf.Tensor>, "y":[<tf.Tensor>, <tf.Tensor>], "z": {"key": <tf.Tensor>}}
    ```

    Args:
        data: A tensor or possibly nested collection of tensors.
        target_type: What kind of tensor(s) to create, either "tf" or "torch".

    Returns:
        A collection with the same structure as `data`, but with any tensors converted to the `target_type`.
    """
    target_instance = {
        "tf": (tf.Tensor, tf.Variable, tf.distribute.DistributedValues), "torch": torch.Tensor, "np": np.ndarray
    }
    conversion_function = {"tf": tf.convert_to_tensor, "torch": torch.from_numpy, "np": np.array}
    if isinstance(data, target_instance[target_type]):
        return data
    elif data is None:
        return None
    elif isinstance(data, dict):
        return {key: to_tensor(value, target_type) for (key, value) in data.items()}
    elif isinstance(data, list):
        return [to_tensor(val, target_type) for val in data]
    elif isinstance(data, tuple):
        return tuple([to_tensor(val, target_type) for val in data])
    elif isinstance(data, set):
        return set([to_tensor(val, target_type) for val in data])
    else:
        return conversion_function[target_type](np.array(data))



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import tempfile
import unittest

import numpy as np
from torch.utils.data.dataset import Dataset as TorchDS

import fastestimator as fe
from fastestimator.dataset import NumpyDataset
from fastestimator.op.numpyop import NumpyOp
from fastestimator.op.tensorop import TensorOp
from fastestimator.op.tensorop.model import ModelOp
from fastestimator.schedule import RepeatScheduler
from fastestimator.test.unittest_util import sample_system_object, sample_system_object_torch
from fastestimator.trace import Trace


class TestNonTraceableDataset(TorchDS):
    def __init__(self, data, var):
        super().__init__()
        self.data = data
        self.var = var

    def __getitem__(self, idx):
        return self.data[idx]

    def __len__(self):
        return len(self.data)


class TestDataset(NumpyDataset):
    def __init__(self, data, var):
        super().__init__(data)
        self.var = var


class TestNumpyOp(NumpyOp):
    def __init__(self, inputs, outputs, mode, var):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        self.var = var


class TestTensorOp(TensorOp):
    def __init__(self, inputs, outputs, mode, var):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        self.var = var


class TestTrace(Trace):
    def __init__(self, var1):
        super().__init__()
        self.var1 = var1


class TestRepeatScheduler(unittest.TestCase):
    def test_save_and_load_state_with_ds_scheduler_tf(self):
        def instantiate_system():
            system = sample_system_object()
            x_train = np.ones((2, 28, 28, 3))
            y_train = np.ones((2, ))

            train_data = RepeatScheduler([
                TestDataset(data={
                    "x": x_train, "y": y_train
                }, var=1),
                TestDataset(data={
                    "x": x_train, "y": y_train
                }, var=1),
            ])
            system.pipeline = fe.Pipeline(train_data=train_data, batch_size=1)
            return system

        system = instantiate_system()

        # make some changes
        new_var = 2
        system.pipeline.data["train"][None].get_current_value(1).var = new_var

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_dir=save_path)

        # reinstantiate system and load the state
        system = instantiate_system()
        system.load_state(save_path)

        loaded_var = system.pipeline.data["train"][None].get_current_value(1).var
        self.assertEqual(loaded_var, new_var)

    def test_save_and_load_state_with_ds_scheduler_torch(self):
        def instantiate_system():
            system = sample_system_object_torch()
            x_train = np.ones((2, 3, 28, 28))
            y_train = np.ones((2, ))

            train_data = RepeatScheduler([
                TestDataset(data={
                    "x": x_train, "y": y_train
                }, var=1),
                TestDataset(data={
                    "x": x_train, "y": y_train
                }, var=1),
            ])
            system.pipeline = fe.Pipeline(train_data=train_data, batch_size=1)
            return system

        system = instantiate_system()

        # make some changes
        new_var = 2
        system.pipeline.data["train"][None].get_current_value(1).var = new_var

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_dir=save_path)

        # reinstantiate system and load the state
        system = instantiate_system()
        system.load_state(save_path)

        loaded_var = system.pipeline.data["train"][None].get_current_value(1).var
        self.assertEqual(loaded_var, new_var)

    def test_save_and_load_state_with_top_scheduler_tf(self):
        def instantiate_system():
            system = sample_system_object()
            model = fe.build(model_fn=fe.architecture.tensorflow.LeNet, optimizer_fn='adam', model_name='tf')
            system.network = fe.Network(ops=[
                RepeatScheduler([
                    TestTensorOp(inputs="x_out", outputs="x_out", mode="train", var=1),
                    TestTensorOp(inputs="x_out", outputs="x_out", mode="train", var=1)
                ]),
                ModelOp(model=model, inputs="x_out", outputs="y_pred")
            ])
            return system

        system = instantiate_system()

        # make some changes
        new_var = 2
        system.network.ops[0].get_current_value(1).var = new_var

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_dir=save_path)

        # reinstantiate system and load the state
        system = instantiate_system()
        system.load_state(save_path)
        loaded_var = system.network.ops[0].get_current_value(1).var

        self.assertEqual(loaded_var, new_var)

    def test_save_and_load_state_with_top_scheduler_torch(self):
        def instantiate_system():
            system = sample_system_object_torch()
            model = fe.build(model_fn=fe.architecture.pytorch.LeNet, optimizer_fn='adam', model_name='torch')
            system.network = fe.Network(ops=[
                RepeatScheduler([
                    TestTensorOp(inputs="x_out", outputs="x_out", mode="train", var=1),
                    TestTensorOp(inputs="x_out", outputs="x_out", mode="train", var=1)
                ]),
                ModelOp(model=model, inputs="x_out", outputs="y_pred")
            ])
            return system

        system = instantiate_system()

        # make some changes
        new_var = 2
        system.network.ops[0].get_current_value(1).var = new_var

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_dir=save_path)

        # reinstantiate system and load the state
        system = instantiate_system()
        system.load_state(save_path)
        loaded_var = system.network.ops[0].get_current_value(1).var

        self.assertEqual(loaded_var, new_var)

    def test_save_and_load_state_with_nop_scheduler_tf(self):
        def instantiate_system():
            system = sample_system_object()
            system.pipeline.ops = [
                RepeatScheduler([
                    TestNumpyOp(inputs="x", outputs="x", mode="train", var=1),
                    TestNumpyOp(inputs="x", outputs="x", mode="train", var=1)
                ])
            ]
            return system

        system = instantiate_system()

        # make some changes
        new_var = 2
        system.pipeline.ops[0].get_current_value(1).var = new_var

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_path)

        # reinstantiate system and load the state
        system = instantiate_system()
        system.load_state(save_path)
        loaded_var = system.pipeline.ops[0].get_current_value(1).var

        self.assertEqual(loaded_var, new_var)

    def test_save_and_load_state_with_nop_scheduler_torch(self):
        def instantiate_system():
            system = sample_system_object_torch()
            system.pipeline.ops = [
                RepeatScheduler([
                    TestNumpyOp(inputs="x", outputs="x", mode="train", var=1),
                    TestNumpyOp(inputs="x", outputs="x", mode="train", var=1)
                ])
            ]
            return system

        system = instantiate_system()

        # make some changes
        new_var = 2
        system.pipeline.ops[0].get_current_value(1).var = new_var

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_path)

        # reinstantiate system and load the state
        system = instantiate_system()
        system.load_state(save_path)
        loaded_var = system.pipeline.ops[0].get_current_value(1).var

        self.assertEqual(loaded_var, new_var)

    def test_save_and_load_state_with_trace_scheduler_tf(self):
        def instantiate_system():
            system = sample_system_object()
            system.traces.append(RepeatScheduler([TestTrace(var1=1), TestTrace(var1=1)]))
            return system

        system = instantiate_system()

        # make some changes
        new_var = 2
        system.traces[0].get_current_value(1).var1 = new_var

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_path)

        # reinstantiate system and load the state
        system = instantiate_system()
        system.load_state(save_path)
        loaded_var = system.traces[0].get_current_value(1).var1

        self.assertEqual(loaded_var, new_var)

    def test_save_and_load_state_with_trace_scheduler_torch(self):
        def instantiate_system():
            system = sample_system_object_torch()
            system.traces.append(RepeatScheduler([TestTrace(var1=1), TestTrace(var1=1)]))
            return system

        system = instantiate_system()

        # make some changes
        new_var = 2
        system.traces[0].get_current_value(1).var1 = new_var

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_path)

        # reinstantiate system and load the state
        system = instantiate_system()
        system.load_state(save_path)
        loaded_var = system.traces[0].get_current_value(1).var1

        self.assertEqual(loaded_var, new_var)

    def test_save_and_load_state_with_non_traceable_entries(self):
        def instantiate_system():
            system = sample_system_object()
            x_train = np.ones((2, 28, 28, 3))
            y_train = np.ones((2, ))
            data = {0: {'x': x_train[0], 'y': y_train[0]}, 1: {'x': x_train[1], 'y': y_train[1]}}
            train_data = RepeatScheduler(
                [TestNonTraceableDataset(data=data, var=3), TestNonTraceableDataset(data=data, var=7)])
            system.pipeline = fe.Pipeline(train_data=train_data, batch_size=1)
            return system

        system = instantiate_system()

        # make some changes
        new_var1 = 4
        system.pipeline.data["train"][None].get_current_value(1).var = new_var1
        new_var2 = 99
        system.pipeline.data["train"][None].get_current_value(2).var = new_var2

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_dir=save_path)

        # reinstantiate system and load the state
        system = instantiate_system()
        system.load_state(save_path)

        with self.subTest('Check that repeat list is still populated'):
            self.assertEqual(2, len(system.pipeline.data["train"][None].repeat_list))
        with self.subTest('Check that classes are still intact'):
            self.assertTrue(isinstance(system.pipeline.data["train"][None].get_current_value(1), TestNonTraceableDataset))
            self.assertTrue(isinstance(system.pipeline.data["train"][None].get_current_value(2), TestNonTraceableDataset))
        with self.subTest('Check that the 1st repeat list entry was not restored'):
            # Since the dataset is not traceable changes shouldn't get restored
            loaded_var = system.pipeline.data["train"][None].get_current_value(1).var
            self.assertEqual(loaded_var, 3)
        with self.subTest('Check that the 2nd repeat list entry was not restored'):
            loaded_var = system.pipeline.data["train"][None].get_current_value(2).var
            self.assertEqual(loaded_var, 7)

    def test_save_and_load_state_with_hybrid_entries(self):
        def instantiate_system():
            system = sample_system_object()
            x_train = np.ones((2, 28, 28, 3))
            y_train = np.ones((2, ))
            data = {0: {'x': x_train[0], 'y': y_train[0]}, 1: {'x': x_train[1], 'y': y_train[1]}}
            train_data = RepeatScheduler([
                TestNonTraceableDataset(data=data, var=3), TestDataset(data={
                    'x': x_train, 'y': y_train
                }, var=7), None
            ])
            system.pipeline = fe.Pipeline(train_data=train_data, batch_size=1)
            return system

        system = instantiate_system()

        # make some changes
        new_var1 = 4
        system.pipeline.data["train"][None].get_current_value(1).var = new_var1
        new_var2 = 99
        system.pipeline.data["train"][None].get_current_value(2).var = new_var2

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_dir=save_path)

        # reinstantiate system and load the state
        system = instantiate_system()
        system.load_state(save_path)

        with self.subTest('Check that repeat list is still populated'):
            self.assertEqual(3, len(system.pipeline.data["train"][None].repeat_list))
        with self.subTest('Check that classes are still intact'):
            self.assertTrue(isinstance(system.pipeline.data["train"][None].get_current_value(1), TestNonTraceableDataset))
            self.assertTrue(isinstance(system.pipeline.data["train"][None].get_current_value(2), TestDataset))
            self.assertTrue(system.pipeline.data["train"][None].get_current_value(3) is None)
        with self.subTest('Check that the 1st repeat list entry was not restored'):
            # Since the dataset is not traceable changes shouldn't get restored
            loaded_var = system.pipeline.data["train"][None].get_current_value(1).var
            self.assertEqual(loaded_var, 3)
        with self.subTest('Check that the 2nd repeat list entry was restored'):
            loaded_var = system.pipeline.data["train"][None].get_current_value(2).var
            self.assertEqual(loaded_var, new_var2)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Sequence, Union, Iterable

import matplotlib.pyplot as plt

from fastestimator.summary.logs.log_plot import visualize_logs
from fastestimator.summary.summary import Summary
from fastestimator.trace.trace import Trace
from fastestimator.util.data import Data
from fastestimator.util.img_data import ImgData
from fastestimator.util.traceability_util import traceable
from fastestimator.util.util import show_image


@traceable()
class ImageViewer(Trace):
    """A trace that interrupts your training in order to display images on the screen.

    This class is useful primarily for Jupyter Notebook, or for debugging purposes.

    Args:
        inputs: Key(s) of images to be displayed.
        mode: What mode(s) to execute this Trace in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        width: The width in inches of the figure.
        height: The height in inches of the figure.
    """
    def __init__(
        self,
        inputs: Union[str, Sequence[str]],
        mode: Union[None, str, Iterable[str]] = ("eval", "test"),
        width: int = 12,
        height: int = 6) -> None:
        super().__init__(inputs=inputs, mode=mode)
        plt.rcParams['figure.figsize'] = [width, height]

    def on_epoch_end(self, data: Data) -> None:
        self._display_images(data)

    def on_end(self, data: Data) -> None:
        self._display_images(data)

    def _display_images(self, data: Data) -> None:
        """A method to render images to the screen.

        Args:
            data: Data possibly containing images to render.
        """
        for key in self.inputs:
            if key in data:
                imgs = data[key]
                if isinstance(imgs, ImgData):
                    fig = imgs.paint_numpy(dpi=96)
                    plt.imshow(fig[0])
                    plt.axis('off')
                    plt.tight_layout()
                    plt.show()
                elif isinstance(imgs, Summary):
                    visualize_logs([imgs])
                elif isinstance(imgs, (list, tuple)) and all([isinstance(img, Summary) for img in imgs]):
                    visualize_logs(imgs)
                else:
                    for idx, img in enumerate(imgs):
                        show_image(img, title="{}_{}".format(key, idx))
                        plt.show()



#  Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ==============================================================================
import statistics as stats
from collections import defaultdict
from typing import Any, Dict, Iterable, Optional, Union

from fastestimator.summary.summary import Summary, ValWithError
from fastestimator.trace.trace import Trace
from fastestimator.util.data import Data
from fastestimator.util.traceability_util import traceable
from fastestimator.util.util import DefaultKeyDict, to_number, to_set


@traceable()
class LabelTracker(Trace):
    """A Trace to track metrics grouped by labels, for example per-class loss over time during training.

    Use this in conjunction with ImageViewer or ImageSaver to see the graph at training end. This also automatically
    integrates with Traceability reports.

    Args:
        label: The key of the labels by which to group data.
        metric: The key of the metric by which to score data.
        label_mapping: A mapping of {DisplayName: LabelValue} to use when generating the graph. This can also be used to
            limit which label values are graphed, since any label values not included here will not be graphed. A None
            value will monitor all label values.
        bounds: What error bounds should be graphed around the mean value. Options include None, 'std' for standard
            deviation, and 'range' to plot (min_value, mean, max_value). Multiple values can be specified, ex.
             ['std', 'range'] to generate multiple graphs.
        mode: What mode(s) to execute this Trace in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Trace in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        outputs: The name of the output which will be generated by this trace at the end of training. If None then it
            will default to "<metric>_by_<label>".

    Raises:
        ValueError: If `bounds` is not one of the allowed options.
    """
    def __init__(self,
                 label: str,
                 metric: str,
                 label_mapping: Optional[Dict[str, Any]] = None,
                 bounds: Union[None, str, Iterable[Union[str, None]]] = "std",
                 mode: Union[None, str, Iterable[str]] = "eval",
                 ds_id: Union[None, str, Iterable[str]] = None,
                 outputs: Optional[str] = None):
        super().__init__(inputs=[label, metric], outputs=outputs or f"{metric}_by_{label}", mode=mode, ds_id=ds_id)
        self.points = []
        self.label_summaries = DefaultKeyDict(default=lambda x: Summary(name=x))
        self.label_mapping = {val: key for key, val in label_mapping.items()} if label_mapping else None
        bounds = to_set(bounds)
        if not bounds:
            bounds.add(None)
        for option in bounds:
            if option not in (None, "std", "range"):
                raise ValueError(f"'interval' must be either None, 'std', or 'range', but got '{bounds}'.")
        self.bounds = bounds

    @property
    def label_key(self) -> str:
        return self.inputs[0]

    @property
    def metric_key(self) -> str:
        return self.inputs[1]

    def on_batch_end(self, data: Data) -> None:
        self.points.append((to_number(data[self.label_key]), to_number(data[self.metric_key])))

    def on_epoch_end(self, data: Data) -> None:
        label_scores = defaultdict(list)
        for batch in self.points:
            for label, metric in ((batch[0][i], batch[1][i]) for i in range(len(batch[0]))):
                label_scores[label.item()].append(metric.item())
        for label, metric in label_scores.items():
            if self.label_mapping:
                if label in self.label_mapping:
                    label = self.label_mapping[label]
                else:
                    # Skip labels which the user does not want to inspect
                    continue
            if 'std' in self.bounds:
                mean, std = stats.mean(metric), stats.stdev(metric) if len(metric) > 1 else 0.0
                val = ValWithError(mean - std, mean, mean + std)
                key = f"{self.metric_key} ($\\mu \\pm \\sigma$)"
                # {label: {mode: {key: {step: value}}}}
                self.label_summaries[label].history[self.system.mode][key][self.system.global_step] = val
            if 'range' in self.bounds:
                val = ValWithError(min(metric), stats.mean(metric), max(metric))
                key = f"{self.metric_key} ($min, \\mu, max$)"
                self.label_summaries[label].history[self.system.mode][key][self.system.global_step] = val
            if None in self.bounds:
                val = stats.mean(metric)
                key = self.metric_key
                self.label_summaries[label].history[self.system.mode][key][self.system.global_step] = val
        self.points = []

    def on_end(self, data: Data) -> None:
        self.system.add_graph(self.outputs[0], list(self.label_summaries.values()))  # So traceability can draw it
        data.write_without_log(self.outputs[0], list(self.label_summaries.values()))

    def __getstate__(self) -> Dict[str, Any]:
        """Get a representation of the state of this object.

        This method is invoked by pickle.

        Returns:
            The information to be recorded by a pickle summary of this object.
        """
        state = self.__dict__.copy()
        state['label_summaries'] = dict(state['label_summaries'])
        return state

    def __setstate__(self, state: Dict[str, Any]) -> None:
        """Set this objects internal state from a dictionary of variables.

        This method is invoked by pickle.

        Args:
            state: The saved state to be used by this object.
        """
        label_summaries = DefaultKeyDict(default=lambda x: Summary(name=x))
        label_summaries.update(state.get('label_summaries', {}))
        state['label_summaries'] = label_summaries
        self.__dict__.update(state)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import tensorflow as tf
import tensorflow_probability as tfp

import fastestimator as fe


class TestInstanceNorm(unittest.TestCase):
    def test_instance_norm(self):
        n = tfp.distributions.Normal(loc=10, scale=2)
        x = n.sample(sample_shape=(1, 100, 100, 1))
        m = fe.layers.tensorflow.InstanceNormalization()
        y = m(x)
        self.assertLess(tf.reduce_mean(y), 0.1)
        self.assertLess(tf.math.reduce_std(y), 0.1)



#  Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ==============================================================================
from collections import defaultdict
from typing import Any, Dict, Iterable, Optional, Set, Union

from fastestimator.summary.summary import Summary
from fastestimator.trace.trace import Trace
from fastestimator.util.data import Data
from fastestimator.util.traceability_util import traceable
from fastestimator.util.util import DefaultKeyDict, to_number, to_set


@traceable()
class InstanceTracker(Trace):
    """A Trace to track metrics by instances, for example per-instance loss over time during training.

    Use this in conjunction with ImageViewer or ImageSaver to see the graph at training end. This also automatically
    integrates with Traceability reports.

    Args:
        index: A key containing data indices to track over time.
        metric: The key of the metric by which to score data.
        n_max_to_keep: At the end of training, the n samples with the highest metric score will be plotted.
        n_min_to_keep: At the end of training, the n samples with the lowest metric score will be plotted.
        list_to_keep: A list of particular indices to pay attention to. This can be used in addition to `n_max_to_keep`
            and/or `n_min_to_keep`, or set those to zero to only track specific indices.
        epoch_frequency: How frequently to collect data. Increase this value to reduce ram consumption.
        mode: What mode(s) to execute this Trace in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Trace in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        outputs: The name of the output which will be generated by this trace at the end of training. If None then it
            will default to "<metric>_by_<index>".

    Raises:
        ValueError: If `n_max_to_keep` or `n_min_to_keep` are invalid.
    """
    def __init__(self,
                 index: str,
                 metric: str,
                 n_max_to_keep: int = 5,
                 n_min_to_keep: int = 5,
                 list_to_keep: Optional[Iterable[Any]] = None,
                 epoch_frequency: int = 1,
                 mode: Union[None, str, Iterable[str]] = "eval",
                 ds_id: Union[None, str, Iterable[str]] = None,
                 outputs: Optional[str] = None):
        # TODO - highlight 'interesting' samples (sudden changes in relative ordering?)
        super().__init__(inputs=[index, metric], outputs=outputs or f"{metric}_by_{index}", mode=mode, ds_id=ds_id)
        self.points = []
        if n_max_to_keep < 0:
            raise ValueError(f"n_max_to_keep must be non-negative, but got {n_max_to_keep}")
        self.n_max_to_keep = n_max_to_keep
        if n_min_to_keep < 0:
            raise ValueError(f"n_min_to_keep must be non-negative, but got {n_min_to_keep}")
        self.n_min_to_keep = n_min_to_keep
        self.idx_to_keep = to_set(list_to_keep)
        # Ideally the step and metric would be separated to save space, but a given idx may not appear each epoch
        self.index_history = defaultdict(lambda: defaultdict(list))  # {mode: {idx: [(step, metric)]}}
        self.epoch_frequency = epoch_frequency

    @property
    def index_key(self) -> str:
        return self.inputs[0]

    @property
    def metric_key(self) -> str:
        return self.inputs[1]

    def on_batch_end(self, data: Data) -> None:
        if self.system.epoch_idx % self.epoch_frequency == 0:
            self.points.append((to_number(data[self.index_key]), to_number(data[self.metric_key])))

    def on_epoch_end(self, data: Data) -> None:
        if self.system.epoch_idx % self.epoch_frequency == 0:
            idx_scores = {}
            for batch in self.points:
                for idx, metric in ((batch[0][i], batch[1][i]) for i in range(len(batch[0]))):
                    idx_scores[idx.item()] = metric.item()
            for idx, metric in idx_scores.items():
                if self.idx_to_keep and self.n_min_to_keep == 0 and self.n_max_to_keep == 0:
                    # We can only skip recording if max_to_keep and min_to_keep are 0 since otherwise we don't know
                    # which histories will need to be thrown out later.
                    if idx not in self.idx_to_keep:
                        # Skip labels which the user does not want to inspect
                        continue
                self.index_history[self.system.mode][idx].append((self.system.global_step, metric))
        self.points = []

    def on_end(self, data: Data) -> None:
        index_summaries = DefaultKeyDict(default=lambda x: Summary(name=x))
        for mode in self.mode:
            final_scores = sorted([(idx, elem[-1][1]) for idx, elem in self.index_history[mode].items()],
                                  key=lambda x: x[1])
            max_idx_list = {elem[0] for elem in final_scores[-1:-self.n_max_to_keep - 1:-1]}
            min_idx_list = {elem[0] for elem in final_scores[:self.n_min_to_keep]}
            target_idx_list = Set.union(min_idx_list, max_idx_list, self.idx_to_keep)
            for idx in target_idx_list:
                for step, score in self.index_history[mode][idx]:
                    index_summaries[idx].history[mode][self.metric_key][step] = score
        self.system.add_graph(self.outputs[0], list(index_summaries.values()))  # So traceability can draw it
        data.write_without_log(self.outputs[0], list(index_summaries.values()))

    def __getstate__(self) -> Dict[str, Any]:
        """Get a representation of the state of this object.

        This method is invoked by pickle.

        Returns:
            The information to be recorded by a pickle summary of this object.
        """
        state = self.__dict__.copy()
        state['index_history'] = dict(state['index_history'])
        return state

    def __setstate__(self, state: Dict[str, Any]) -> None:
        """Set this objects internal state from a dictionary of variables.

        This method is invoked by pickle.

        Args:
            state: The saved state to be used by this object.
        """
        index_history = defaultdict(lambda: defaultdict(list))
        index_history.update(state.get('index_history', {}))
        state['index_history'] = index_history
        self.__dict__.update(state)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, Dict, Generic, Iterable, List, Optional, TypeVar, Union

from fastestimator.util.traceability_util import is_restorable, traceable
from fastestimator.util.util import to_set

T = TypeVar('T')


@traceable()
class Scheduler(Generic[T]):
    """A class which can wrap things like Datasets and Ops to make their behavior epoch-dependent.
    """
    def get_current_value(self, epoch: int) -> Optional[T]:
        """Fetch whichever of the `Scheduler`s elements is appropriate based on the current epoch.

        Args:
            epoch: The current epoch.

        Returns:
            The element from the Scheduler to be used at the given `epoch`. This value might be None.
        """
        raise NotImplementedError

    def get_all_values(self) -> List[Optional[T]]:
        """Get a list of all the possible values stored in the `Scheduler`.

        Returns:
            A list of all the values stored in the `Scheduler`. This may contain None values.
        """
        raise NotImplementedError


@traceable()
class RepeatScheduler(Scheduler[T]):
    """A scheduler which repeats a collection of entries one after another every epoch.

    One case where this class would be useful is if you want to perform one version of an Op on even epochs, and a
    different version on odd epochs. None values can be used to achieve an end result of skipping an Op every so often.

    ```python
    s = fe.schedule.RepeatScheduler(["a", "b", "c"])
    s.get_current_value(epoch=1)  # "a"
    s.get_current_value(epoch=2)  # "b"
    s.get_current_value(epoch=3)  # "c"
    s.get_current_value(epoch=4)  # "a"
    s.get_current_value(epoch=5)  # "b"
    ```

    Args:
        repeat_list: What elements to cycle between every epoch. Note that epochs start counting from 1. To have nothing
        happen for a particular epoch, None values may be used.

    Raises:
        AssertionError: If `repeat_list` is not a List.
    """
    def __init__(self, repeat_list: List[Optional[T]]) -> None:
        assert isinstance(repeat_list, List), "must provide a list as input of RepeatSchedule"
        self.repeat_list = repeat_list
        self.cycle_length = len(repeat_list)
        assert self.cycle_length > 1, "list length must be greater than 1"

    def get_current_value(self, epoch: int) -> Optional[T]:
        # epoch-1 since the training epoch is 1-indexed rather than 0-indexed.
        return self.repeat_list[(epoch - 1) % self.cycle_length]

    def get_all_values(self) -> List[Optional[T]]:
        return self.repeat_list

    def __getstate__(self) -> Dict[str, List[Dict[Any, Any]]]:
        return {
            'repeat_list': [
                elem if is_restorable(elem)[0] else elem.__getstate__() if hasattr(elem, '__getstate__') else {}
                for elem in self.repeat_list
            ]
        }


@traceable()
class EpochScheduler(Scheduler[T]):
    """A scheduler which selects entries based on a specified epoch mapping.

    This can be useful for making networks grow over time, or to use more challenging data augmentation as training
    progresses.

    ```python
    s = fe.schedule.EpochScheduler({1:"a", 3:"b", 4:None, 100: "c"})
    s.get_current_value(epoch=1)  # "a"
    s.get_current_value(epoch=2)  # "a"
    s.get_current_value(epoch=3)  # "b"
    s.get_current_value(epoch=4)  # None
    s.get_current_value(epoch=99)  # None
    s.get_current_value(epoch=100)  # "c"
    ```

    Args:
        epoch_dict: A mapping from epoch -> element. For epochs in between keys in the dictionary, the closest prior key
            will be used to determine which element to return. None values may be used to cause nothing to happen for a
            particular epoch.

    Raises:
        AssertionError: If the `epoch_dict` is of the wrong type, or contains invalid keys.
    """
    def __init__(self, epoch_dict: Dict[int, T]) -> None:
        assert isinstance(epoch_dict, dict), "must provide dictionary as epoch_dict"
        self.epoch_dict = epoch_dict
        self.keys = sorted(self.epoch_dict)
        for key in self.keys:
            assert isinstance(key, int), "found non-integer key: {}".format(key)
            assert key >= 1, "found non-positive key: {}".format(key)

    def get_current_value(self, epoch: int) -> Optional[T]:
        if epoch in self.keys:
            value = self.epoch_dict[epoch]
        else:
            last_key = self._get_last_key(epoch)
            if last_key is None:
                value = None
            else:
                value = self.epoch_dict[last_key]
        return value

    def get_all_values(self) -> List[Optional[T]]:
        return list(self.epoch_dict.values())

    def _get_last_key(self, epoch: int) -> Union[int, None]:
        """Find the nearest prior key to the given epoch.

        Args:
            epoch: The current target epoch.

        Returns:
            The largest epoch number <= the given `epoch` that is in the `epoch_dict`.
        """
        last_key = None
        for key in self.keys:
            if key > epoch:
                break
            last_key = key
        return last_key

    def __getstate__(self) -> Dict[str, Dict[int, Dict[Any, Any]]]:
        return {
            'epoch_dict': {
                key: elem if is_restorable(elem)[0] else elem.__getstate__() if hasattr(elem, '__getstate__') else {}
                for key,
                elem in self.epoch_dict.items()
            }
        }


def get_signature_epochs(items: List[Any], total_epochs: int, mode: Optional[str] = None) -> List[int]:
    """Find all epochs of changes due to schedulers.

    Args:
        items: List of items to scan from.
        total_epochs: The maximum epoch number to consider when searching for signature epochs.
        mode: Current execution mode. If None, all execution modes will be considered.

    Returns:
        The epoch numbers of changes.
    """
    unique_configs = []
    signature_epochs = []
    for epoch in range(1, total_epochs + 1):
        epoch_config = get_current_items(items, run_modes=mode, epoch=epoch)
        if epoch_config not in unique_configs:
            unique_configs.append(epoch_config)
            signature_epochs.append(epoch)
    return signature_epochs


def get_current_items(items: Iterable[Union[T, Scheduler[T]]],
                      run_modes: Optional[Union[str, Iterable[str]]] = None,
                      epoch: Optional[int] = None,
                      ds_id: Optional[str] = None) -> List[T]:
    """Select items which should be executed for given mode and epoch.

    Args:
        items: A list of possible items or Schedulers of items to choose from.
        run_modes: The desired execution mode. One or more of "train", "eval", "test", or "infer". If None, items of
            all modes will be returned.
        epoch: The desired execution epoch. If None, items across all epochs will be returned.
        ds_id: The desired execution dataset id. If None, items across all ds_ids will be returned.

    Returns:
        The items which should be executed.
    """
    selected_items = []
    run_modes = to_set(run_modes)
    for item in items:
        if isinstance(item, Scheduler):
            if epoch is None:
                item = item.get_all_values()
            else:
                item = [item.get_current_value(epoch)]
        else:
            item = [item]
        for item_ in item:
            # mode matching
            mode_match = False
            if not run_modes:
                mode_match = True
            if not hasattr(item_, "mode"):
                mode_match = True
            else:
                if not item_.mode:
                    mode_match = True
                elif item_.mode.intersection(run_modes):
                    mode_match = True

            # ds_id matching
            ds_id_match = False
            if not ds_id:
                ds_id_match = True
            if not hasattr(item_, "ds_id"):
                ds_id_match = True
            else:
                if not item_.ds_id:
                    ds_id_match = True
                elif ds_id in item_.ds_id:
                    ds_id_match = True  # whitelist check
                # if any of ds_id starts with "!", then they will all start with "!"
                elif any([x.startswith("!") for x in item_.ds_id]) and all([ds_id != x[1:] for x in item_.ds_id]):
                    ds_id_match = True  # blacklist check
            if item_ and mode_match and ds_id_match:
                selected_items.append(item_)
    return selected_items



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import tensorflow as tf
import torch

from fastestimator.op.tensorop.augmentation.mixup_batch import MixUpBatch
from fastestimator.test.unittest_util import is_equal


class MyTFBeta:
    @staticmethod
    def sample(sample_shape=(1, )):
        return 0.5 * tf.ones(shape=sample_shape)


class MyTorchBeta:
    @staticmethod
    def sample(sample_shape=(1, )):
        return 0.5 * torch.ones(size=sample_shape)


class TestMixUpBatch(unittest.TestCase):
    def test_tf_shared_beta(self):
        data = tf.constant([[1.0], [3.0], [4.5]])
        expected = tf.constant([[2.75], [2], [3.75]])
        mu = MixUpBatch(inputs="x", outputs=["x", "lambda"], alpha=1.0, mode="train", shared_beta=True)
        mu.build('tf')
        mu.beta = MyTFBeta()
        output = mu.forward(data=[data], state={})
        self.assertTrue(is_equal(output[0], expected))

    def test_tf_different_beta(self):
        data = tf.constant([[1.0], [3.0], [4.5]])
        expected = tf.constant([[2.75], [2], [3.75]])
        mu = MixUpBatch(inputs="x", outputs=["x", "lambda"], alpha=1.0, mode="train", shared_beta=False)
        mu.build('tf')
        mu.beta = MyTFBeta()
        output = mu.forward(data=[data], state={})
        self.assertTrue(is_equal(output[0], expected))

    def test_torch_shared_beta(self):
        data = torch.tensor([[1.0], [3.0], [4.5]])
        expected = torch.tensor([[2.75], [2], [3.75]])
        mu = MixUpBatch(inputs="x", outputs=["x", "lambda"], alpha=1.0, mode="train", shared_beta=True)
        mu.build('torch')
        mu.beta = MyTorchBeta()
        output = mu.forward(data=[data], state={})
        self.assertTrue(is_equal(output[0], expected))

    def test_torch_different_beta(self):
        data = torch.tensor([[1.0], [3.0], [4.5]])
        expected = torch.tensor([[2.75], [2], [3.75]])
        mu = MixUpBatch(inputs="x", outputs=["x", "lambda"], alpha=1.0, mode="train", shared_beta=False)
        mu.build('torch')
        mu.beta = MyTorchBeta()
        output = mu.forward(data=[data], state={})
        self.assertTrue(is_equal(output[0], expected))



# Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Iterable, List, Optional, Tuple, Union

from albumentations.imgaug.transforms import IAACropAndPad as IAACropAndPadAlb

from fastestimator.op.numpyop.multivariate.multivariate import MultiVariateAlbumentation
from fastestimator.util.traceability_util import traceable


@traceable()
class IAACropAndPad(MultiVariateAlbumentation):
    """Crop and pad images by pixel amounts or fractions of image sizes.

    Args:
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        image_in: The key of an image to be modified.
        mask_in: The key of a mask to be modified (with the same random factors as the image).
        masks_in: The key of masks to be modified (with the same random factors as the image).
        image_out: The key to write the modified image (defaults to `image_in` if None).
        mask_out: The key to write the modified mask (defaults to `mask_in` if None).
        masks_out: The key to write the modified masks (defaults to `masks_in` if None).
        px: he number of pixels to crop (negative values) or pad (positive values) on each side of the image.
            Either this or the parameter percent may be set, not both at the same time.
        percent: The number of pixels to crop (negative values) or pad (positive values) on each side of the image given
             as a fraction of the image height/width.
        pad_mode: OpenCV border mode.
        pad_cval: The constant value to use if the pad mode is BORDER_CONSTANT.  If a tuple of two number s and at least
             one of them is a float, then a random number will be uniformly sampled per image from the continuous
             interval [a, b] and used as the value. If both number s are int s, the interval is discrete. If a list of
             number, then a random value will be chosen from the elements of the list and used as the value.
        keep_size: Whether to keep the same size as original image.

    Image types:
        uint8, float32
    """
    def __init__(self,
                 px: Union[None, int, Tuple[int, int]] = None,
                 percent: Union[None, float, Tuple[float, float]] = None,
                 pad_mode: Union[int, str] = 'constant',
                 pad_cval: Union[int, Tuple[float], List[int]] = 0,
                 keep_size: bool = True,
                 mode: Union[None, str, Iterable[str]] = None,
                 ds_id: Union[None, str, Iterable[str]] = None,
                 image_in: Optional[str] = None,
                 mask_in: Optional[str] = None,
                 masks_in: Optional[str] = None,
                 image_out: Optional[str] = None,
                 mask_out: Optional[str] = None,
                 masks_out: Optional[str] = None):
        super().__init__(
            IAACropAndPadAlb(px=px,
                             percent=percent,
                             pad_mode=pad_mode,
                             pad_cval=pad_cval,
                             keep_size=keep_size,
                             always_apply=True),
            image_in=image_in,
            mask_in=mask_in,
            masks_in=masks_in,
            bbox_in=None,
            keypoints_in=None,
            image_out=image_out,
            mask_out=mask_out,
            masks_out=masks_out,
            bbox_out=None,
            keypoints_out=None,
            bbox_params=None,
            keypoint_params=None,
            mode=mode,
            ds_id=ds_id)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Optional, TypeVar

import tensorflow as tf
import torch

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor)


def watch(tensor: Tensor, tape: Optional[tf.GradientTape] = None) -> Tensor:
    """Monitor the given `tensor` for later gradient computations.

    This method can be used with TensorFlow tensors:
    ```python
    x = tf.ones((3,28,28,1))
    with tf.GradientTape(persistent=True) as tape:
        x = fe.backend.watch(x, tape=tape)
    ```

    This method can be used with PyTorch tensors:
    ```python
    x = torch.ones((3,1,28,28))  # x.requires_grad == False
    x = fe.backend.watch(x)  # x.requires_grad == True
    ```

    Args:
        tensor: The tensor to be monitored.
        tape: A TensorFlow GradientTape which will be used to record gradients (iff using TensorFlow for the backend).

    Returns:
        The `tensor` or a copy of the `tensor` which is being tracked for gradient computations. This value is only
        needed if using PyTorch as the backend.

    Raises:
        ValueError: If `tensor` is an unacceptable data type.
    """
    if tf.is_tensor(tensor):
        tape.watch(tensor)
        return tensor
    elif isinstance(tensor, torch.Tensor):
        if tensor.requires_grad:
            return tensor
        # It is tempting to just do tensor.requires_grad = True here, but that will lead to trouble
        return tensor.detach().requires_grad_(True)
    else:
        raise ValueError("Unrecognized tensor type {}".format(type(tensor)))



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf
import torch

from fastestimator.backend import random_normal_like


class TestRandomNormalLike(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_np = np.array([[0, 1], [1, 1]])
        cls.test_tf = tf.constant([[0, 1], [2, 2]])
        cls.test_torch = torch.Tensor([[1, 1], [2, 3]])

    def test_random_normal_np_type(self):
        self.assertIsInstance(random_normal_like(self.test_np), np.ndarray, 'Output must be NumPy Array')

    def test_random_normal_np_value(self):
        self.assertTrue((random_normal_like(self.test_np).shape == (2, 2)),
                        'Output array shape should be same as input')

    def test_random_normal_tf_type(self):
        self.assertIsInstance(random_normal_like(self.test_tf), tf.Tensor, 'Output type must be tf.Tensor')

    def test_random_normal_tf_value(self):
        output_shape = tf.shape([2, 2])
        self.assertTrue(tf.reduce_all(tf.equal(tf.shape(random_normal_like(self.test_tf)), output_shape)),
                        'Output tensor shape should be same as input')

    def test_random_normal_torch_type(self):
        self.assertIsInstance(random_normal_like(self.test_torch), torch.Tensor, 'Output must be torch.Tensor')

    def test_random_normal_torch_value(self):
        output_shape = (2, 2)
        self.assertTrue((random_normal_like(self.test_torch).size() == output_shape),
                        'Output tensor shape should be same as input')



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import tempfile
import unittest

import dill
import numpy as np

from fastestimator.test.unittest_util import is_equal
from fastestimator.test.unittest_util import sample_system_object
from fastestimator.trace.adapt import PBMCalibrator
from fastestimator.util.data import Data


class TestPBMCalibrator(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        save_dir = tempfile.mkdtemp()
        cls.save_file = os.path.join(save_dir, 'calibrator.pkl')
        cls.pbm_calibrator = PBMCalibrator(true_key='y', pred_key='y_pred', save_path=cls.save_file)
        cls.pbm_calibrator.system = sample_system_object()

    def test_on_epoch_begin(self):
        self.pbm_calibrator.on_epoch_begin(data=Data())
        with self.subTest('Check initial value of y_true'):
            self.assertEqual(self.pbm_calibrator.y_true, [])
        with self.subTest('Check initial value of y_pred'):
            self.assertEqual(self.pbm_calibrator.y_pred, [])

    def test_on_batch_end(self):
        self.pbm_calibrator.y_true = []
        self.pbm_calibrator.y_pred = []
        batch1 = {'y': np.array([0, 0, 1, 1]), 'y_pred': np.array([[1.0, 0.0], [1.0, 0.0], [0.0, 1.0], [0.0, 1.0]])}
        self.pbm_calibrator.on_batch_end(data=Data(batch1))
        with self.subTest('Check true values'):
            self.assertTrue(is_equal(self.pbm_calibrator.y_true, list(batch1['y'])))
        with self.subTest('Check pred values'):
            self.assertTrue(is_equal(self.pbm_calibrator.y_pred, list(batch1['y_pred'])))
        batch2 = {'y': np.array([1, 1, 0, 0]), 'y_pred': np.array([[0.0, 1.0], [0.0, 1.0], [1.0, 0.0], [1.0, 0.0]])}
        self.pbm_calibrator.on_batch_end(data=Data(batch2))
        with self.subTest('Check true values (2 batches)'):
            self.assertTrue(is_equal(self.pbm_calibrator.y_true, list(batch1['y']) + list(batch2['y'])))
        with self.subTest('Check pred values (2 batches)'):
            self.assertTrue(is_equal(self.pbm_calibrator.y_pred, list(batch1['y_pred']) + list(batch2['y_pred'])))

    def test_on_epoch_end(self):
        self.pbm_calibrator.y_true = [0] * 50 + [1] * 50
        self.pbm_calibrator.y_pred = list(np.array([1.0, 0.0] * 50 + [0.0, 1.0] * 50).reshape(100, 2))
        expected = np.array([1.0, 0.0] * 50 + [0.0, 1.0] * 50).reshape(100, 2)
        data = Data()
        self.pbm_calibrator.on_epoch_end(data=data)
        with self.subTest('Check if output exists'):
            self.assertIn('y_pred_calibrated', data)
        with self.subTest('Check save file exists'):
            self.assertTrue(os.path.exists(self.save_file))
        with self.subTest('Check the calibrated values'):
            self.assertTrue(np.allclose(data['y_pred_calibrated'], expected))
        with self.subTest('Check the save file performance'):
            with open(self.save_file, 'rb') as f:
                fn = dill.load(f)
                resp = fn(expected)
                self.assertTrue(np.array_equal(resp, data['y_pred_calibrated']))



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import TypeVar

import numpy as np
import tensorflow as tf
import torch

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor, np.ndarray)


def expand_dims(tensor: Tensor, axis: int = 1) -> Tensor:
    """Create a new dimension in `tensor` along a given `axis`.

    This method can be used with Numpy data:
    ```python
    n = np.array([2,7,5])
    b = fe.backend.expand_dims(n, axis=0)  # [[2, 5, 7]]
    b = fe.backend.expand_dims(n, axis=1)  # [[2], [5], [7]]
    ```

    This method can be used with TensorFlow tensors:
    ```python
    t = tf.constant([2,7,5])
    b = fe.backend.expand_dims(t, axis=0)  # [[2, 5, 7]]
    b = fe.backend.expand_dims(t, axis=1)  # [[2], [5], [7]]
    ```

    This method can be used with PyTorch tensors:
    ```python
    p = torch.tensor([2,7,5])
    b = fe.backend.expand_dims(p, axis=0)  # [[2, 5, 7]]
    b = fe.backend.expand_dims(p, axis=1)  # [[2], [5], [7]]
    ```

    Args:
        tensor: The input to be modified, having n dimensions.
        axis: Which axis should the new axis be inserted along. Must be in the range [-n-1, n].

    Returns:
        A concatenated representation of the `tensors`, or None if the list of `tensors` was empty.

    Raises:
        ValueError: If `tensor` is an unacceptable data type.
    """
    if tf.is_tensor(tensor):
        return tf.expand_dims(tensor, axis=axis)
    elif isinstance(tensor, torch.Tensor):
        return torch.unsqueeze(tensor, dim=axis)
    elif isinstance(tensor, np.ndarray):
        return np.expand_dims(tensor, axis=axis)
    else:
        raise ValueError("Unrecognized tensor type {}".format(type(tensor)))



import tempfile
import unittest

import fastestimator as fe
from fastestimator.op.tensorop import TensorOp
from fastestimator.op.tensorop.model import ModelOp
from fastestimator.test.unittest_util import sample_system_object, sample_system_object_torch


class TestTensorOp(TensorOp):
    def __init__(self, inputs, outputs, mode, var):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode)
        self.var = var


class TestFuse(unittest.TestCase):
    def test_save_and_load_state_tf(self):
        def instantiate_system():
            system = sample_system_object()
            model = fe.build(model_fn=fe.architecture.tensorflow.LeNet, optimizer_fn='adam', model_name='tf')
            system.network = fe.Network(ops=[
                fe.op.tensorop.meta.Fuse(ops=[
                    TestTensorOp(inputs="x_out", outputs="x_out", mode="train", var=1),
                    TestTensorOp(inputs="x_out", outputs="x_out", mode="train", var=1)
                ]),
                ModelOp(model=model, inputs="x_out", outputs="y_pred")
            ])
            return system

        system = instantiate_system()

        # make some changes
        new_var = 2
        system.network.ops[0].ops[0].var = new_var

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_path)

        # reinstantiate system and load the state
        system = instantiate_system()
        system.load_state(save_path)
        loaded_var = system.network.ops[0].ops[0].var

        self.assertEqual(loaded_var, new_var)

    def test_save_and_load_state_torch(self):
        def instantiate_system():
            system = sample_system_object_torch()
            model = fe.build(model_fn=fe.architecture.pytorch.LeNet, optimizer_fn='adam', model_name='torch')
            system.network = fe.Network(ops=[
                fe.op.tensorop.meta.Fuse(ops=[
                    TestTensorOp(inputs="x_out", outputs="x_out", mode="train", var=1),
                    TestTensorOp(inputs="x_out", outputs="x_out", mode="train", var=1)
                ]),
                ModelOp(model=model, inputs="x_out", outputs="y_pred")
            ])
            return system

        system = instantiate_system()

        # make some changes
        new_var = 2
        system.network.ops[0].ops[0].var = new_var

        # save the state
        save_path = tempfile.mkdtemp()
        system.save_state(save_path)

        # reinstantiate system and load the state
        system = instantiate_system()
        system.load_state(save_path)
        loaded_var = system.network.ops[0].ops[0].var

        self.assertEqual(loaded_var, new_var)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from copy import deepcopy
from typing import Any, Dict, Iterable, List, Optional, Sequence, Set, Tuple

import numpy as np

from fastestimator.dataset.dataset import DatasetSummary
from fastestimator.dataset.labeled_dir_dataset import LabeledDirDataset
from fastestimator.util.traceability_util import traceable


@traceable()
class SiameseDirDataset(LabeledDirDataset):
    """A dataset which returns pairs of data.

    This dataset reads files from a folder hierarchy like root/class(/es)/data.file. Data is returned in pairs,
    where the label value is 1 if the data are drawn from the same class, and 0 otherwise. One epoch is defined as
    the time it takes to visit every data point exactly once as the 'data_key_left'. Each data point may occur zero
    or many times as 'data_key_right' within the same epoch. SiameseDirDataset.split() will split by class index
    rather than by data instance index.

    Args:
        root_dir: The path to the directory containing data sorted by folders.
        data_key_left: What key to assign to the first data element in the pair.
        data_key_right: What key to assign to the second data element in the pair.
        label_key: What key to assign to the label values in the data dictionary.
        percent_matching_data: What percentage of the time should data be paired by class (label value = 1).
        label_mapping: A dictionary defining the mapping to use. If not provided will map classes to int labels.
        file_extension: If provided then only files ending with the file_extension will be included.
    """

    class_data: Dict[Any, Set[int]]

    def __init__(self,
                 root_dir: str,
                 data_key_left: str = "x_a",
                 data_key_right: str = "x_b",
                 label_key: str = "y",
                 percent_matching_data: float = 0.5,
                 label_mapping: Optional[Dict[str, Any]] = None,
                 file_extension: Optional[str] = None):
        super().__init__(root_dir, data_key_left, label_key, label_mapping, file_extension)
        self.class_data = self._data_to_class(self.data, label_key)
        self.percent_matching_data = percent_matching_data
        self.data_key_left = data_key_left
        self.data_key_right = data_key_right
        self.label_key = label_key

    @staticmethod
    def _data_to_class(data: Dict[int, Dict[str, Any]], label_key: str) -> Dict[Any, Set[int]]:
        """A helper method to build a mapping from classes to their corresponding data indices.

        Args:
            data: A data dictionary like {<index>: {"data_key": <data value>}}.
            label_key: Which key inside `data` corresponds to the label value for a given index entry.

        Returns:
            A mapping like {"label1": {<indices with label1>}, "label2": {<indices with label2>}}.
        """
        class_data = {}
        for idx, elem in data.items():
            class_data.setdefault(elem[label_key], set()).add(idx)
        return class_data

    def _split_length(self) -> int:
        """The length of a dataset to be used for the purpose of computing splits.

        In this case, splits are computed based on the number of classes rather than the number of instances per class.

        Returns:
            The apparent length of the dataset for the purpose of the .split() function
        """
        return len(self.class_data)

    def _do_split(self, splits: Sequence[Iterable[int]]) -> List['SiameseDirDataset']:
        """Split the current dataset apart into several smaller datasets.

        Args:
            splits: Which classes to remove from the current dataset in order to create new dataset(s). One dataset will
                be generated for every iterable within the `splits` sequence.

        Returns:
            New datasets generated by removing classes at the indices specified by `splits` from the current dataset.
        """
        # Splits in this context refer to class indices rather than the typical data indices
        results = []
        for split in splits:
            # Convert class indices to data indices
            int_class_keys = list(sorted(self.class_data.keys()))
            split = [item for i in split for item in self.class_data[int_class_keys[i]]]
            data = {new_idx: self.data.pop(old_idx) for new_idx, old_idx in enumerate(split)}
            class_data = self._data_to_class(data, self.label_key)
            results.append(
                self._skip_init(data,
                                class_data=class_data,
                                **{k: v
                                   for k, v in self.__dict__.items() if k not in {'data', 'class_data'}}))
        # Re-key the remaining data to be contiguous from 0 to new max index
        self.data = {new_idx: v for new_idx, (old_idx, v) in enumerate(self.data.items())}
        self.class_data = self._data_to_class(self.data, self.label_key)
        # The summary function is being cached by a base class, so reset our cache here
        # noinspection PyUnresolvedReferences
        self.summary.cache_clear()
        return results

    def _get_stratified_splits(self, split_counts: List[int], seed: Optional[int],
                               stratify: str) -> Sequence[Iterable[int]]:
        print("\033[93m {}\033[00m".format(
            "Warning! SiameseDirDataset does not support stratified splits. Falling back to classical split method."))
        return self._get_fractional_splits(split_counts=split_counts, seed=seed)

    def __getitem__(self, index: int):
        """Extract items from the dataset based on the given `batch_idx`.

        Args:
            index: Which data instance to use as the 'left' element.

        Returns:
            A datapoint for the given index.
        """
        base_item = deepcopy(self.data[index])
        if np.random.uniform(0, 1) < self.percent_matching_data:
            # Generate matching data
            clazz_items = self.class_data[base_item[self.label_key]]
            other = np.random.choice(list(clazz_items - {index}))
            base_item[self.data_key_right] = self.data[other][self.data_key_left]
            base_item[self.label_key] = 1
        else:
            # Generate non-matching data
            other_classes = self.class_data.keys() - {base_item[self.label_key]}
            other_class = np.random.choice(list(other_classes))
            other = np.random.choice(list(self.class_data[other_class]))
            base_item[self.data_key_right] = self.data[other][self.data_key_left]
            base_item[self.label_key] = 0
        return base_item

    def one_shot_trial(self, n: int) -> Tuple[List[str], List[str]]:
        """Generate one-shot trial data.

        The similarity should be highest between the index 0 elements of the arrays.

        Args:
            n: The number of samples to draw for computing one shot accuracy. Should be <= the number of total classes.

        Returns:
            ([class_a_instance_x, class_a_instance_x, class_a_instance_x, ...],
            [class_a_instance_w, class_b_instance_y, class_c_instance_z, ...])
        """
        assert n > 1, "one_shot_trial requires an n-value of at least 2"
        assert n <= len(self.class_data.keys()), \
            "one_shot_trial only supports up to {} comparisons, but an n-value of {} was given".format(
                len(self.class_data.keys()), n)
        classes = np.random.choice(list(self.class_data.keys()), size=n, replace=False)
        base_image_indices = np.random.choice(list(self.class_data[classes[0]]), size=2, replace=False)
        l1 = [self.data[base_image_indices[0]][self.data_key_left]] * n
        l2 = [self.data[base_image_indices[1]][self.data_key_left]]
        for clazz in classes[1:]:
            index = np.random.choice(list(self.class_data[clazz]))
            l2.append(self.data[index][self.data_key_left])
        return l1, l2

    def summary(self) -> DatasetSummary:
        """Generate a summary representation of this dataset.
        Returns:
            A summary representation of this dataset.
        """
        summary = super().summary()
        # since class key is re-mapped, remove class key mapping to reduce confusion
        summary.class_key_mapping = None
        return summary



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
from pathlib import Path
from typing import List, Optional, Tuple

import numpy as np
import wget

from fastestimator.dataset.numpy_dataset import NumpyDataset
from fastestimator.util.wget_util import bar_custom, callback_progress

wget.callback_progress = callback_progress


def load_data(root_dir: Optional[str] = None,
              seq_length: int = 64) -> Tuple[NumpyDataset, NumpyDataset, NumpyDataset, List[str]]:
    """Load and return the Penn TreeBank dataset.

    Args:
        root_dir: The path to store the downloaded data. When `path` is not provided, the data will be saved into
            `fastestimator_data` under the user's home directory.
        seq_length: Length of data sequence.

    Returns:
        (train_data, eval_data, test_data, vocab)
    """
    home = str(Path.home())

    if root_dir is None:
        root_dir = os.path.join(home, 'fastestimator_data', 'PennTreeBank')
    else:
        root_dir = os.path.join(os.path.abspath(root_dir), 'PennTreeBank')
    os.makedirs(root_dir, exist_ok=True)

    train_data_path = os.path.join(root_dir, 'ptb.train.txt')
    eval_data_path = os.path.join(root_dir, 'ptb.valid.txt')
    test_data_path = os.path.join(root_dir, 'ptb.test.txt')

    files = [(train_data_path, 'https://raw.githubusercontent.com/wojzaremba/lstm/master/data/ptb.train.txt'),
             (eval_data_path, 'https://raw.githubusercontent.com/wojzaremba/lstm/master/data/ptb.valid.txt'),
             (test_data_path, 'https://raw.githubusercontent.com/wojzaremba/lstm/master/data/ptb.test.txt')]

    texts = []
    for data_path, download_link in files:
        if not os.path.exists(data_path):
            # Download
            print("Downloading data: {}".format(data_path))
            wget.download(download_link, data_path, bar=bar_custom)

        text = []
        with open(data_path, 'r') as f:
            for line in f:
                text.extend(line.split() + ['<eos>'])
        texts.append(text)

    # Build dictionary from training data
    vocab = sorted(set(texts[0]))
    word2idx = {u: i for i, u in enumerate(vocab)}

    #convert word to index and split the sequences and discard the last incomplete sequence
    data = [[word2idx[word] for word in text[:-(len(text) % seq_length)]] for text in texts]
    x_train, x_eval, x_test = [np.array(d).reshape(-1, seq_length) for d in data]

    train_data = NumpyDataset(data={"x": x_train})
    eval_data = NumpyDataset(data={"x": x_eval})
    test_data = NumpyDataset(data={"x": x_test})
    return train_data, eval_data, test_data, vocab



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import re
from collections import defaultdict
from typing import List, Optional, Set, Iterable

from fastestimator.dataset.dir_dataset import DirDataset
from fastestimator.summary.logs.log_plot import visualize_logs
from fastestimator.summary.summary import Summary, ValWithError, average_summaries
from fastestimator.util.util import strip_suffix


def parse_log_iter(source: Iterable[str], sync: Summary) -> Summary:
    """A function which will parse lines into a dictionary of metrics.

    Args:
        source: A collection of lines to parse.
        sync: The summary to append into.

    Returns:
        The updated summary object.
    """
    last_step = 0
    last_epoch = 0
    for line in source:
        mode = None
        if line.startswith("FastEstimator-Train") or line.startswith("FastEstimator-Finish"):
            mode = "train"
        elif line.startswith("FastEstimator-Eval"):
            mode = "eval"
        elif line.startswith("FastEstimator-Test"):
            mode = "test"
        if mode is None:
            continue
        num = r"([-]?[0-9]+[.]?[0-9]*(e[-]?[0-9]+[.]?[0-9]*)?)"
        parsed_line = re.findall(r"([^:;]+):[\s]*(" + num + r"|None|\(" + num + ", " + num + ", " + num + r"\));", line)
        step = parsed_line[0]
        assert step[0].strip() == "step", \
            "Log file (%s) seems to be missing step information, or step is not listed first" % sync.name
        step = step[1]
        adjust_epoch = False
        if step == 'None':
            # This might happen if someone runs the test mode from the cli
            step = last_step
            # If the test mode was just guessing its epoch, use the prior epoch instead
            adjust_epoch = mode == 'test'
        else:
            step = int(step)
            last_step = step
        for metric in parsed_line[1:]:
            if metric[4]:
                val = ValWithError(float(metric[4]), float(metric[6]), float(metric[8]))
            else:
                val = metric[1]
                if val == 'None':
                    continue
                val = float(val)
            key = metric[0].strip()
            if key == 'epoch':
                if adjust_epoch:
                    val = last_epoch
                else:
                    last_epoch = val
            sync.history[mode][key].update({step: val})
    return sync


def parse_log_file(file_path: str, file_extension: str) -> Summary:
    """A function which will parse log files into a dictionary of metrics.

    Args:
        file_path: The path to a log file.
        file_extension: The extension of the log file.
    Returns:
        An experiment summarizing the given log file.
    """
    # TODO: need to handle multi-line output like confusion matrix
    experiment = Summary(strip_suffix(os.path.split(file_path)[1].strip(), file_extension))
    with open(file_path) as file:
        parse_log_iter(source=file, sync=experiment)
    return experiment


def parse_log_files(file_paths: List[str],
                    log_extension: Optional[str] = '.txt',
                    smooth_factor: float = 0,
                    save: bool = False,
                    save_path: Optional[str] = None,
                    ignore_metrics: Optional[Set[str]] = None,
                    include_metrics: Optional[Set[str]] = None,
                    share_legend: bool = True,
                    pretty_names: bool = False,
                    group_by: Optional[str] = None) -> None:
    """Parse one or more log files for graphing.

    This function which will iterate through the given log file paths, parse them to extract metrics, remove any
    metrics which are blacklisted, and then pass the necessary information on the graphing function.

    Args:
        file_paths: A list of paths to various log files.
        log_extension: The extension of the log files.
        smooth_factor: A non-negative float representing the magnitude of gaussian smoothing to apply (zero for none).
        save: Whether to save (True) or display (False) the generated graph.
        save_path: Where to save the image if save is true. Defaults to dir_path if not provided.
        ignore_metrics: Any metrics within the log files which will not be visualized.
        include_metrics: A whitelist of metric keys (None whitelists all keys).
        share_legend: Whether to have one legend across all graphs (True) or one legend per graph (False).
        pretty_names: Whether to modify the metric names in graph titles (True) or leave them alone (False).
        group_by: Combine multiple log files by a regex to visualize their mean+-stddev. For example, to group together
            files like [a_1.txt, a_2.txt] vs [b_1.txt, b_2.txt] you can use: r'(.*)_[\d]+\.txt'.

    Raises:
        AssertionError: If no log files are provided.
        ValueError: If a log file does not match the `group_by` regex pattern.
    """
    if file_paths is None or len(file_paths) < 1:
        raise AssertionError("must provide at least one log file")
    if save and save_path is None:
        save_path = file_paths[0]

    groups = defaultdict(list)  # {group_name: [experiment(s)]}
    for path in file_paths:
        experiment = parse_log_file(path, log_extension)
        try:
            key = (re.findall(group_by, os.path.split(path)[1]))[0] if group_by else experiment.name
        except IndexError:
            raise ValueError(f"The log {os.path.split(path)[1]} did not match the given regex pattern: {group_by}")
        groups[key].append(experiment)
    experiments = [average_summaries(name, exps) for name, exps in groups.items()]

    visualize_logs(experiments,
                   save_path=save_path,
                   smooth_factor=smooth_factor,
                   share_legend=share_legend,
                   pretty_names=pretty_names,
                   ignore_metrics=ignore_metrics,
                   include_metrics=include_metrics)


def parse_log_dir(dir_path: str,
                  log_extension: str = '.txt',
                  recursive_search: bool = False,
                  smooth_factor: float = 1,
                  save: bool = False,
                  save_path: Optional[str] = None,
                  ignore_metrics: Optional[Set[str]] = None,
                  include_metrics: Optional[Set[str]] = None,
                  share_legend: bool = True,
                  pretty_names: bool = False,
                  group_by: Optional[str] = None) -> None:
    """A function which will gather all log files within a given folder and pass them along for visualization.

    Args:
        dir_path: The path to a directory containing log files.
        log_extension: The extension of the log files.
        recursive_search: Whether to recursively search sub-directories for log files.
        smooth_factor: A non-negative float representing the magnitude of gaussian smoothing to apply (zero for none).
        save: Whether to save (True) or display (False) the generated graph.
        save_path: Where to save the image if save is true. Defaults to dir_path if not provided.
        ignore_metrics: Any metrics within the log files which will not be visualized.
        include_metrics: A whitelist of metric keys (None whitelists all keys).
        share_legend: Whether to have one legend across all graphs (True) or one legend per graph (False).
        pretty_names: Whether to modify the metric names in graph titles (True) or leave them alone (False).
        group_by: Combine multiple log files by a regex to visualize their mean+-stddev. For example, to group together
            files like [a_1.txt, a_2.txt] vs [b_1.txt, b_2.txt] you can use: r'(.*)_[\d]+\.txt'.
    """
    loader = DirDataset(root_dir=dir_path, file_extension=log_extension, recursive_search=recursive_search)
    file_paths = list(map(lambda d: d['x'], loader.data.values()))

    parse_log_files(file_paths,
                    log_extension,
                    smooth_factor,
                    save,
                    save_path,
                    ignore_metrics,
                    include_metrics,
                    share_legend,
                    pretty_names,
                    group_by)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import unittest

import tensorflow as tf
import torch

from fastestimator.op.tensorop.augmentation import CutMixBatch
from fastestimator.test.unittest_util import MockBetaDistribution, MockUniformDistribution, check_img_similar, \
    img_to_rgb_array


class TestCutMixBatch(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_tf_bbox = tf.random.uniform((4, 32, 32, 3))
        cls.tf_x = tf.constant([0.25])
        cls.tf_y = tf.constant([0.25])
        cls.tf_lam = tf.constant([0.5])
        cls.tf_output_bbox = (tf.constant([0]), tf.constant([19]), tf.constant([0]), tf.constant([19]), 32, 32)
        cls.test_torch_bbox = torch.rand((4, 3, 32, 32))
        cls.torch_x = torch.tensor([0.25])
        cls.torch_y = torch.tensor([0.25])
        cls.torch_lam = torch.tensor([0.5])
        cls.torch_output_bbox = (torch.tensor([0]), torch.tensor([19]), torch.tensor([0]), torch.tensor([19]), 32, 32)
        cls.tf_input = tf.stack([tf.ones((32, 32, 3)), tf.zeros((32, 32, 3))])
        cls.torch_input = torch.cat([torch.ones((1, 3, 32, 32)), torch.zeros((1, 3, 32, 32))], dim=0)
        cls.cutmix_output1 = os.path.abspath(
            os.path.join(__file__, "..", "..", "..", "..", "util", "resources", "test_cutmix1.png"))
        cls.cutmix_output2 = os.path.abspath(
            os.path.join(__file__, "..", "..", "..", "..", "util", "resources", "test_cutmix2.png"))

    def test_tf_bbox_coordinates(self):
        cutmix = CutMixBatch(inputs='x', outputs=['x', 'lam'])
        output = cutmix._get_patch_coordinates(tensor=self.test_tf_bbox, x=self.tf_x, y=self.tf_y, lam=self.tf_lam)
        with self.subTest('Check length of output tuple'):
            self.assertEqual(len(output), 6)
        with self.subTest('Check output value'):
            self.assertEqual(output, self.tf_output_bbox)

    def test_torch_bbox_coordinates(self):
        cutmix = CutMixBatch(inputs='x', outputs=['x', 'lam'])
        output = cutmix._get_patch_coordinates(tensor=self.test_torch_bbox,
                                               x=self.torch_x,
                                               y=self.torch_y,
                                               lam=self.torch_lam)
        with self.subTest('Check length of output tuple'):
            self.assertEqual(len(output), 6)
        with self.subTest('Check output value'):
            self.assertEqual(output, self.torch_output_bbox)

    def test_tf_output(self):
        cutmix = CutMixBatch(inputs='x', outputs=['x', 'lam'])
        cutmix.beta = MockBetaDistribution('tf')
        cutmix.uniform = MockUniformDistribution('tf')
        mixed_images = cutmix.forward(data=self.tf_input, state={})
        images = mixed_images[0].numpy()
        lam = mixed_images[1].numpy()
        with self.subTest('First mixed image'):
            self.assertTrue(check_img_similar(images[0], img_to_rgb_array(self.cutmix_output1)))
        with self.subTest('Second mixed image'):
            self.assertTrue(check_img_similar(images[1], img_to_rgb_array(self.cutmix_output2)))
        with self.subTest('lambda value'):
            self.assertEqual(round(float(lam), 2), 0.65)

    def test_torch_output(self):
        cutmix = CutMixBatch(inputs='x', outputs=['x', 'lam'])
        cutmix.beta = MockBetaDistribution('torch')
        cutmix.uniform = MockUniformDistribution('torch')
        mixed_images = cutmix.forward(data=self.torch_input, state={})
        images = mixed_images[0].permute(0, 2, 3, 1)
        images = images.numpy()
        lam = mixed_images[1].numpy()
        with self.subTest('First mixed image'):
            self.assertTrue(check_img_similar(images[0], img_to_rgb_array(self.cutmix_output1)))
        with self.subTest('Second mixed image'):
            self.assertTrue(check_img_similar(images[1], img_to_rgb_array(self.cutmix_output2)))
        with self.subTest('lambda value'):
            self.assertEqual(round(float(lam), 2), 0.65)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf

import fastestimator as fe


class TestNumpyDataset(unittest.TestCase):
    def test_dataset(self):
        (x_train, y_train), _ = tf.keras.datasets.mnist.load_data()
        train_data = fe.dataset.NumpyDataset({"x": x_train, "y": y_train})

        self.assertEqual(len(train_data), 60000)

    def test_single_frac_split(self):
        ds1 = fe.dataset.NumpyDataset(
            {"idx": np.array([i for i in range(100)]), "clz": np.array([i % 4 for i in range(100)])})
        ds2 = ds1.split(0.3)
        with self.subTest("Source Dataset Size"):
            self.assertEqual(len(ds1), 70)
        with self.subTest("New Dataset Size"):
            self.assertEqual(len(ds2), 30)
        with self.subTest("Disjoint datasets"):
            self.assertEqual(set(), set(ds1["idx"]) & set(ds2["idx"]))

    def test_single_frac_split_seed(self):
        ds1 = fe.dataset.NumpyDataset(
            {"idx": np.array([i for i in range(100)]), "clz": np.array([i % 4 for i in range(100)])})
        ds2 = ds1.split(0.3, seed=42)
        with self.subTest("Source Dataset Size"):
            self.assertEqual(len(ds1), 70)
        with self.subTest("New Dataset Size"):
            self.assertEqual(len(ds2), 30)
        with self.subTest("Disjoint datasets"):
            self.assertEqual(set(), set(ds1["idx"]) & set(ds2["idx"]))

        ds3 = fe.dataset.NumpyDataset(
            {"idx": np.array([i for i in range(100)]), "clz": np.array([i % 4 for i in range(100)])})
        ds4 = ds3.split(0.3, seed=42)
        with self.subTest("Both source datasets should be equivalent"):
            self.assertEqual(ds1["idx"], ds3["idx"])
        with self.subTest("Both new datasets should be equivalent"):
            self.assertEqual(ds2["idx"], ds4["idx"])

    def test_single_frac_split_stratify(self):
        ds1 = fe.dataset.NumpyDataset(
            {"idx": np.array([i for i in range(100)]), "clz": np.array([i % 4 for i in range(100)])})
        ds2 = ds1.split(0.4, stratify="clz")
        with self.subTest("Source Dataset Size"):
            self.assertEqual(len(ds1), 60)
        with self.subTest("New Dataset Size"):
            self.assertEqual(len(ds2), 40)
        with self.subTest("Disjoint datasets"):
            self.assertEqual(set(), set(ds1["idx"]) & set(ds2["idx"]))
        with self.subTest("Source class balance maintained"):
            self.assertEqual(ds1["clz"].count(0), 15)
            self.assertEqual(ds1["clz"].count(1), 15)
            self.assertEqual(ds1["clz"].count(2), 15)
            self.assertEqual(ds1["clz"].count(3), 15)
        with self.subTest("New class balance maintained"):
            self.assertEqual(ds2["clz"].count(0), 10)
            self.assertEqual(ds2["clz"].count(1), 10)
            self.assertEqual(ds2["clz"].count(2), 10)
            self.assertEqual(ds2["clz"].count(3), 10)

    def test_single_frac_split_seed_stratify(self):
        ds1 = fe.dataset.NumpyDataset(
            {"idx": np.array([i for i in range(100)]), "clz": np.array([i % 4 for i in range(100)])})
        ds2 = ds1.split(0.4, seed=17, stratify="clz")
        with self.subTest("Source Dataset Size"):
            self.assertEqual(len(ds1), 60)
        with self.subTest("New Dataset Size"):
            self.assertEqual(len(ds2), 40)
        with self.subTest("Disjoint datasets"):
            self.assertEqual(set(), set(ds1["idx"]) & set(ds2["idx"]))
        with self.subTest("Source class balance maintained"):
            self.assertEqual(ds1["clz"].count(0), 15)
            self.assertEqual(ds1["clz"].count(1), 15)
            self.assertEqual(ds1["clz"].count(2), 15)
            self.assertEqual(ds1["clz"].count(3), 15)
        with self.subTest("New class balance maintained"):
            self.assertEqual(ds2["clz"].count(0), 10)
            self.assertEqual(ds2["clz"].count(1), 10)
            self.assertEqual(ds2["clz"].count(2), 10)
            self.assertEqual(ds2["clz"].count(3), 10)

        ds3 = fe.dataset.NumpyDataset(
            {"idx": np.array([i for i in range(100)]), "clz": np.array([i % 4 for i in range(100)])})
        ds4 = ds3.split(0.4, seed=17, stratify="clz")
        with self.subTest("Both source datasets should be equivalent"):
            self.assertEqual(ds1["idx"], ds3["idx"])
        with self.subTest("Both new datasets should be equivalent"):
            self.assertEqual(ds2["idx"], ds4["idx"])

    def test_multi_frac_split(self):
        ds1 = fe.dataset.NumpyDataset(
            {"idx": np.array([i for i in range(100)]), "clz": np.array([i % 4 for i in range(100)])})
        ds2, ds3 = ds1.split(0.3, 0.4)
        with self.subTest("Source Dataset Size"):
            self.assertEqual(len(ds1), 30)
        with self.subTest("New Dataset Size"):
            self.assertEqual(len(ds2), 30)
            self.assertEqual(len(ds3), 40)
        with self.subTest("Disjoint datasets"):
            self.assertEqual(set(), set(ds1["idx"]) & set(ds2["idx"]))
            self.assertEqual(set(), set(ds2["idx"]) & set(ds3["idx"]))
            self.assertEqual(set(), set(ds1["idx"]) & set(ds3["idx"]))

    def test_multi_frac_split_seed(self):
        ds1 = fe.dataset.NumpyDataset(
            {"idx": np.array([i for i in range(100)]), "clz": np.array([i % 4 for i in range(100)])})
        ds2, ds3 = ds1.split(0.3, 0.4, seed=42)
        with self.subTest("Source Dataset Size"):
            self.assertEqual(len(ds1), 30)
        with self.subTest("New Dataset Size"):
            self.assertEqual(len(ds2), 30)
            self.assertEqual(len(ds3), 40)
        with self.subTest("Disjoint datasets"):
            self.assertEqual(set(), set(ds1["idx"]) & set(ds2["idx"]))
            self.assertEqual(set(), set(ds2["idx"]) & set(ds3["idx"]))
            self.assertEqual(set(), set(ds1["idx"]) & set(ds3["idx"]))

        ds4 = fe.dataset.NumpyDataset(
            {"idx": np.array([i for i in range(100)]), "clz": np.array([i % 4 for i in range(100)])})
        ds5, ds6 = ds4.split(0.3, 0.4, seed=42)
        with self.subTest("Both source datasets should be equivalent"):
            self.assertEqual(ds1["idx"], ds4["idx"])
        with self.subTest("New datasets should be equivalent"):
            self.assertEqual(ds2["idx"], ds5["idx"])
            self.assertEqual(ds3["idx"], ds6["idx"])

    def test_multi_frac_split_stratify(self):
        ds1 = fe.dataset.NumpyDataset(
            {"idx": np.array([i for i in range(100)]), "clz": np.array([i % 4 for i in range(100)])})
        ds2, ds3 = ds1.split(0.4, 0.2, stratify="clz")
        with self.subTest("Source Dataset Size"):
            self.assertEqual(len(ds1), 40)
        with self.subTest("New Dataset Size"):
            self.assertEqual(len(ds2), 40)
            self.assertEqual(len(ds3), 20)
        with self.subTest("Disjoint datasets"):
            self.assertEqual(set(), set(ds1["idx"]) & set(ds2["idx"]))
            self.assertEqual(set(), set(ds2["idx"]) & set(ds3["idx"]))
            self.assertEqual(set(), set(ds1["idx"]) & set(ds3["idx"]))
        with self.subTest("Source class balance maintained"):
            self.assertEqual(ds1["clz"].count(0), 10)
            self.assertEqual(ds1["clz"].count(1), 10)
            self.assertEqual(ds1["clz"].count(2), 10)
            self.assertEqual(ds1["clz"].count(3), 10)
        with self.subTest("New class balance maintained"):
            self.assertEqual(ds2["clz"].count(0), 10)
            self.assertEqual(ds2["clz"].count(1), 10)
            self.assertEqual(ds2["clz"].count(2), 10)
            self.assertEqual(ds2["clz"].count(3), 10)
            self.assertEqual(ds3["clz"].count(0), 5)
            self.assertEqual(ds3["clz"].count(1), 5)
            self.assertEqual(ds3["clz"].count(2), 5)
            self.assertEqual(ds3["clz"].count(3), 5)

    def test_multi_frac_split_seed_stratify(self):
        ds1 = fe.dataset.NumpyDataset(
            {"idx": np.array([i for i in range(100)]), "clz": np.array([i % 4 for i in range(100)])})
        ds2, ds3 = ds1.split(0.4, 0.2, seed=17, stratify="clz")
        with self.subTest("Source Dataset Size"):
            self.assertEqual(len(ds1), 40)
        with self.subTest("New Dataset Size"):
            self.assertEqual(len(ds2), 40)
            self.assertEqual(len(ds3), 20)
        with self.subTest("Disjoint datasets"):
            self.assertEqual(set(), set(ds1["idx"]) & set(ds2["idx"]))
            self.assertEqual(set(), set(ds2["idx"]) & set(ds3["idx"]))
            self.assertEqual(set(), set(ds1["idx"]) & set(ds3["idx"]))
        with self.subTest("Source class balance maintained"):
            self.assertEqual(ds1["clz"].count(0), 10)
            self.assertEqual(ds1["clz"].count(1), 10)
            self.assertEqual(ds1["clz"].count(2), 10)
            self.assertEqual(ds1["clz"].count(3), 10)
        with self.subTest("New class balance maintained"):
            self.assertEqual(ds2["clz"].count(0), 10)
            self.assertEqual(ds2["clz"].count(1), 10)
            self.assertEqual(ds2["clz"].count(2), 10)
            self.assertEqual(ds2["clz"].count(3), 10)
            self.assertEqual(ds3["clz"].count(0), 5)
            self.assertEqual(ds3["clz"].count(1), 5)
            self.assertEqual(ds3["clz"].count(2), 5)
            self.assertEqual(ds3["clz"].count(3), 5)

        ds4 = fe.dataset.NumpyDataset(
            {"idx": np.array([i for i in range(100)]), "clz": np.array([i % 4 for i in range(100)])})
        ds5, ds6 = ds4.split(0.4, 0.2, seed=17, stratify="clz")
        with self.subTest("Both source datasets should be equivalent"):
            self.assertEqual(ds1["idx"], ds4["idx"])
        with self.subTest("New datasets should be equivalent"):
            self.assertEqual(ds2["idx"], ds5["idx"])
            self.assertEqual(ds3["idx"], ds6["idx"])

    def test_heavily_imbalanced_stratify_split(self):
        idx_array = np.array([i for i in range(100)])
        clz_array = np.array(
            [0 if i < 10 else 1 if i < 40 else 2 if i < 77 else 3 if i < 84 else 4 if i < 97 else 5 if i < 99 else 6 for
             i in range(100)])
        # 0: 10%, 1: 30%, 2: 37%, 3: 7%, 4: 13%, 5: 2%, 6: 1%
        ds1 = fe.dataset.NumpyDataset(
            {"idx": idx_array, "clz": clz_array})
        ds2, ds3 = ds1.split(0.3, 0.25, stratify="clz")
        with self.subTest("Source Dataset Size"):
            self.assertEqual(len(ds1), 45)
        with self.subTest("New Dataset Size"):
            self.assertEqual(len(ds2), 30)
            self.assertEqual(len(ds3), 25)
        with self.subTest("Disjoint datasets"):
            self.assertEqual(set(), set(ds1["idx"]) & set(ds2["idx"]))
            self.assertEqual(set(), set(ds2["idx"]) & set(ds3["idx"]))
            self.assertEqual(set(), set(ds1["idx"]) & set(ds3["idx"]))
        with self.subTest("DS1 class balance maintained"):
            self.assertEqual(ds1["clz"].count(0), 5)  # 11.11%
            self.assertEqual(ds1["clz"].count(1), 13)  # 28.88%
            self.assertEqual(ds1["clz"].count(2), 18)  # 40%
            self.assertEqual(ds1["clz"].count(3), 3)  # 6.66%
            self.assertEqual(ds1["clz"].count(4), 6)  # 13.33%
            self.assertEqual(ds1["clz"].count(5), 0)  # 0%
            self.assertEqual(ds1["clz"].count(6), 0)  # 0%
        with self.subTest("DS2 class balance maintained"):
            self.assertEqual(ds2["clz"].count(0), 3)  # 10%
            self.assertEqual(ds2["clz"].count(1), 9)  # 30%
            self.assertEqual(ds2["clz"].count(2), 10)  # 33.33%
            self.assertEqual(ds2["clz"].count(3), 2)  # 6.66%
            self.assertEqual(ds2["clz"].count(4), 4)  # 13.33%
            self.assertEqual(ds2["clz"].count(5), 1)  # 3.33%
            self.assertEqual(ds2["clz"].count(6), 1)  # 3.33%
        with self.subTest("DS3 class balance maintained"):
            self.assertEqual(ds3["clz"].count(0), 2)  # 8%
            self.assertEqual(ds3["clz"].count(1), 8)  # 32%
            self.assertEqual(ds3["clz"].count(2), 9)  # 36%
            self.assertEqual(ds3["clz"].count(3), 2)  # 8%
            self.assertEqual(ds3["clz"].count(4), 3)  # 12%
            self.assertEqual(ds3["clz"].count(5), 1)  # 4%
            self.assertEqual(ds3["clz"].count(6), 0)  # 0%



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Any, Dict, List, Optional, Set, TypeVar

import tensorflow as tf
import tensorflow_probability as tfp
import torch

from fastestimator.op.tensorop.tensorop import TensorOp
from fastestimator.util.traceability_util import traceable

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor)
Model = TypeVar('Model', tf.keras.Model, torch.nn.Module)


@traceable()
class Sometimes(TensorOp):
    """Perform a NumpyOp with a given probability.

    Note that Sometimes should not be used to wrap an op whose output key(s) do not already exist in the data
    dictionary. This would result in a problem when future ops / traces attempt to reference the output key, but
    Sometimes declined to generate it. If you want to create a default value for a new key, simply use a LambdaOp before
    invoking the Sometimes.

    Args:
        tensor_op: The operator to be performed.
        prob: The probability of execution, which should be in the range: [0-1).
    """
    def __init__(self, tensor_op: TensorOp, prob: float = 0.5) -> None:
        # We're going to try to collect any missing output keys from the data dictionary so that they don't get
        # overridden when Sometimes chooses not to execute.
        inps = set(tensor_op.inputs)
        outs = set(tensor_op.outputs)
        self.extra_inputs = list(outs - inps)  # Used by traceability
        self.inp_idx = len(tensor_op.inputs)
        super().__init__(inputs=tensor_op.inputs + self.extra_inputs,
                         outputs=tensor_op.outputs,
                         mode=tensor_op.mode,
                         ds_id=tensor_op.ds_id)
        # Note that in_list and out_list will always be true
        self.op = tensor_op
        self.prob = prob
        self.prob_fn = None

    def build(self, framework: str, device: Optional[torch.device] = None) -> None:
        self.op.build(framework, device)
        if framework == 'tf':
            self.prob_fn = tfp.distributions.Uniform()
        elif framework == 'torch':
            self.prob_fn = torch.distributions.uniform.Uniform(low=0, high=1)
        else:
            raise ValueError("unrecognized framework: {}".format(framework))

    def get_fe_loss_keys(self) -> Set[str]:
        return self.op.get_fe_loss_keys()

    def get_fe_models(self) -> Set[Model]:
        return self.op.get_fe_models()

    def fe_retain_graph(self, retain: Optional[bool] = None) -> Optional[bool]:
        return self.op.fe_retain_graph(retain)

    def __getstate__(self) -> Dict[str, Dict[Any, Any]]:
        return {'op': self.op.__getstate__() if hasattr(self.op, '__getstate__') else {}}

    def forward(self, data: List[Tensor], state: Dict[str, Any]) -> List[Tensor]:
        """Execute the wrapped operator a certain fraction of the time.

        Args:
            data: The information to be passed to the wrapped operator.
            state: Information about the current execution context, for example {"mode": "train"}.

        Returns:
            The original `data`, or the `data` after running it through the wrapped operator.
        """
        if self.prob > self.prob_fn.sample():
            data = data[:self.inp_idx]  # Cut off the unnecessary inputs
            if not self.op.in_list:
                data = data[0]
            data = self.op.forward(data, state)
            if not self.op.out_list:
                data = [data]
        else:
            data = [data[self.inputs.index(out)] for out in self.outputs]
        return data



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np
import tensorflow as tf
import torch

import fastestimator as fe
from fastestimator.test.unittest_util import is_equal


class TestGatherFromBatch(unittest.TestCase):
    def test_gather_from_batch_np_input(self):
        tensor = np.array([[0, 1], [2, 3], [4, 5]])
        indice = np.array([1, 0, 1])
        obj1 = fe.backend.gather_from_batch(tensor, indice)
        obj2 = np.array([1, 2, 5])
        self.assertTrue(is_equal(obj1, obj2))

    def test_gather_from_batch_tf_input(self):
        tensor = tf.constant([[0, 1], [2, 3], [4, 5]])
        indice = tf.constant([1, 0, 1])
        obj1 = fe.backend.gather_from_batch(tensor, indice)
        obj2 = tf.constant([1, 2, 5])
        self.assertTrue(is_equal(obj1, obj2))

        tensor = tf.constant([[[0, 1], [2, 3]], [[4, 5], [6, 7]], [[8, 9], [10, 11]]])
        obj1 = fe.backend.gather_from_batch(tensor, indice)
        obj2 = tf.constant([[2, 3], [4, 5], [10, 11]])
        self.assertTrue(is_equal(obj1, obj2))

    def test_gather_from_batch_torch_input_2D(self):
        tensor = torch.tensor([[0, 1], [2, 3], [4, 5]])
        indice = torch.tensor([1, 0, 1])
        obj1 = fe.backend.gather_from_batch(tensor, indice)
        obj2 = torch.tensor([1, 2, 5])
        self.assertTrue(is_equal(obj1, obj2))

    def test_gather_from_batch_torch_input_3D(self):
        tensor = torch.tensor([[[0, 1], [2, 3]], [[4, 5], [6, 7]], [[8, 9], [10, 11]]])
        indice = torch.tensor([1, 0, 1])
        obj1 = fe.backend.gather_from_batch(tensor, indice)
        obj2 = torch.tensor([[2, 3], [4, 5], [10, 11]])
        self.assertTrue(is_equal(obj1, obj2))

    def test_gather_from_batch_input_and_indice_diff_type_2D(self):
        tensor = torch.tensor([[0, 1], [2, 3], [4, 5]])
        indice = np.array([1, 0, 1])
        obj1 = fe.backend.gather_from_batch(tensor, indice)
        obj2 = torch.tensor([1, 2, 5])
        self.assertTrue(is_equal(obj1, obj2))

    def test_gather_from_batch_input_and_indice_diff_type_3D(self):
        tensor = tf.constant([[[0, 1], [2, 3]], [[4, 5], [6, 7]], [[8, 9], [10, 11]]])
        indice = torch.tensor([1, 0, 1])
        obj1 = fe.backend.gather_from_batch(tensor, indice)
        obj2 = tf.constant([[2, 3], [4, 5], [10, 11]])
        self.assertTrue(is_equal(obj1, obj2))



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
import tempfile
import unittest

from fastestimator.test.unittest_util import sample_system_object
from fastestimator.trace.io import RestoreWizard
from fastestimator.util.data import Data


def get_model_name(system):
    model_names = []
    for model in system.network.models:
        model_names.append(model.model_name)
    return model_names


class TestRestoreWizard(unittest.TestCase):
    def test_save(self):
        save_path = tempfile.mkdtemp()
        restore_wizard = RestoreWizard(directory=save_path)
        restore_wizard.system = sample_system_object()
        restore_wizard.on_begin(Data())
        restore_wizard.on_epoch_end(Data())
        with self.subTest("Check Saved Files (1)"):
            self.assertTrue(os.path.exists(os.path.join(save_path, 'key.txt')))
            self.assertTrue(os.path.exists(os.path.join(save_path, 'A')))
        with self.subTest("Check Key is Correct (1)"):
            with open(os.path.join(save_path, 'key.txt'), 'r') as file:
                key = file.readline()
                self.assertEqual(key, "A")
        restore_wizard.on_epoch_end(Data())
        with self.subTest("Check Saved Files (2)"):
            self.assertTrue(os.path.exists(os.path.join(save_path, 'key.txt')))
            self.assertTrue(os.path.exists(os.path.join(save_path, 'B')))
        with self.subTest("Check Key is Correct (2)"):
            with open(os.path.join(save_path, 'key.txt'), 'r') as file:
                key = file.readline()
                self.assertEqual(key, "B")
        restore_wizard.on_epoch_end(Data())
        with self.subTest("Check Saved Files (3)"):
            self.assertTrue(os.path.exists(os.path.join(save_path, 'key.txt')))
            self.assertTrue(os.path.exists(os.path.join(save_path, 'A')))
        with self.subTest("Check Key is Correct (3)"):
            with open(os.path.join(save_path, 'key.txt'), 'r') as file:
                key = file.readline()
                self.assertEqual(key, "A")
        restore_wizard.on_epoch_end(Data())
        with self.subTest("Check Saved Files (4)"):
            self.assertTrue(os.path.exists(os.path.join(save_path, 'key.txt')))
            self.assertTrue(os.path.exists(os.path.join(save_path, 'B')))
        with self.subTest("Check Key is Correct (4)"):
            with open(os.path.join(save_path, 'key.txt'), 'r') as file:
                key = file.readline()
                self.assertEqual(key, "B")

    def test_restore(self):
        save_path = tempfile.mkdtemp()
        global_step = 100
        epoch_idx = 10

        restore_wizard = RestoreWizard(directory=save_path)
        restore_wizard.system = sample_system_object()
        restore_wizard.on_begin(Data())
        restore_wizard.system.global_step = global_step
        restore_wizard.system.epoch_idx = epoch_idx
        restore_wizard.on_epoch_end(Data())

        restore_wizard = RestoreWizard(directory=save_path)
        restore_wizard.system = sample_system_object()
        data = Data()
        restore_wizard.on_begin(data)
        with self.subTest("Check print message"):
            self.assertEqual(data['epoch'], 10)
        with self.subTest("Check system variables"):
            self.assertEqual(restore_wizard.system.global_step, global_step)
            self.assertEqual(restore_wizard.system.epoch_idx, epoch_idx)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np

from fastestimator.test.unittest_util import sample_system_object
from fastestimator.trace.metric import MeanAveragePrecision
from fastestimator.util import Data


class TestMeanAveragePrecision(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        x = np.random.rand(1, 5, 5)
        x_pred = np.random.rand(1, 5, 7)
        cls.data = Data({'x': x, 'x_pred': x_pred})
        cls.map = MeanAveragePrecision(true_key='x', pred_key='x_pred', num_classes=3)
        cls.iou_element_shape = (5, 5)
        cls.map.system = sample_system_object()

    def test_on_epoch_begin(self):
        self.map.on_epoch_begin(data=self.data)
        with self.subTest('Check initial value of image ids'):
            self.assertEqual(self.map.image_ids, [])
        with self.subTest('Check initial value of eval images'):
            self.assertEqual(self.map.evalimgs, {})
        with self.subTest('Check initial value of eval'):
            self.assertEqual(self.map.eval, {})
        with self.subTest('Check initial value of ids in epoch'):
            self.assertEqual(self.map.ids_in_epoch, 0)

    def test_reshape_gt(self):
        x = np.random.rand(1, 5, 5)
        output = self.map._reshape_gt(x)
        self.assertEqual(output.shape, (5, 6))

    def test_reshape_pred(self):
        x = np.random.rand(2, 5, 7)
        output = self.map._reshape_pred(x)
        self.assertEqual(output.shape, (10, 7))

    def test_on_batch_end(self):
        self.map.on_batch_end(data=self.data)
        with self.subTest('Check IOU element shape'):
            self.assertEqual(self.map.ious[(1, 0)].shape, (5, 5))

    def test_on_epoch_end(self):
        self.map.ious = {(1, 0): np.random.uniform(low=0, high=1, size=(5, 5)), (1, 1): [], (1, 2): []}
        self.evalimgs = {
            (0, 1): {
                'image_id': 1,
                'category_id': 0,
                'gtIds': [1, 1, 1, 1, 1],
                'dtMatches': np.random.randint(2, size=(10, 5)),
                'gtMatches': np.random.randint(2, size=(10, 5)),
                'dtScores': np.random.uniform(low=0, high=1, size=(5, )),
                'num_gt': 5
            }, (1, 1): None, (2, 1): None
        }
        self.map.on_epoch_end(data=self.data)
        with self.subTest('Check if mAP exists'):
            self.assertIn('mAP', self.data)
        with self.subTest('Check if AP50 exists'):
            self.assertIn('AP50', self.data)
        with self.subTest('Check if AP75 exists'):
            self.assertIn('AP75', self.data)
        with self.subTest('Check the value of mAP'):
            self.assertEqual(self.data['mAP'], -1)
        with self.subTest('Check the value of AP50'):
            self.assertEqual(self.data['AP50'], -1)
        with self.subTest('Check the value of AP75'):
            self.assertEqual(self.data['AP75'], -1)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
# see a complete list of effects in albumentation page: https://albumentations-demo.herokuapp.com
from fastestimator.op.numpyop.multivariate.affine import Affine
from fastestimator.op.numpyop.multivariate.center_crop import CenterCrop
from fastestimator.op.numpyop.multivariate.crop import Crop
from fastestimator.op.numpyop.multivariate.crop_non_empty_mask_if_exists import CropNonEmptyMaskIfExists
from fastestimator.op.numpyop.multivariate.elastic_transform import ElasticTransform
from fastestimator.op.numpyop.multivariate.flip import Flip
from fastestimator.op.numpyop.multivariate.grid_distortion import GridDistortion
from fastestimator.op.numpyop.multivariate.horizontal_flip import HorizontalFlip
from fastestimator.op.numpyop.multivariate.iaa_crop_and_pad import IAACropAndPad
from fastestimator.op.numpyop.multivariate.longest_max_size import LongestMaxSize
from fastestimator.op.numpyop.multivariate.mask_dropout import MaskDropout
from fastestimator.op.numpyop.multivariate.optical_distortion import OpticalDistortion
from fastestimator.op.numpyop.multivariate.pad_if_needed import PadIfNeeded
from fastestimator.op.numpyop.multivariate.random_crop import RandomCrop
from fastestimator.op.numpyop.multivariate.random_crop_near_bbox import RandomCropNearBBox
from fastestimator.op.numpyop.multivariate.random_grid_shuffle import RandomGridShuffle
from fastestimator.op.numpyop.multivariate.random_resized_crop import RandomResizedCrop
from fastestimator.op.numpyop.multivariate.random_rotate_90 import RandomRotate90
from fastestimator.op.numpyop.multivariate.random_scale import RandomScale
from fastestimator.op.numpyop.multivariate.random_sized_bbox_safe_crop import RandomSizedBBoxSafeCrop
from fastestimator.op.numpyop.multivariate.random_sized_crop import RandomSizedCrop
from fastestimator.op.numpyop.multivariate.read_mat import ReadMat
from fastestimator.op.numpyop.multivariate.resize import Resize
from fastestimator.op.numpyop.multivariate.rotate import Rotate
from fastestimator.op.numpyop.multivariate.shift_scale_rotate import ShiftScaleRotate
from fastestimator.op.numpyop.multivariate.smallest_max_size import SmallestMaxSize
from fastestimator.op.numpyop.multivariate.transpose import Transpose
from fastestimator.op.numpyop.multivariate.vertical_flip import VerticalFlip



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Iterable, Union

from fastestimator.op.tensorop.tensorop import TensorOp


class LossOp(TensorOp):
    """Abstract base LossOp class.

    A base class for loss operations. It can be used directly to perform value pass-through (see the adversarial
    training showcase for an example of when this is useful).

    Args:
        inputs: A tuple or list like: [<y_pred>, <y_true>].
        outputs: String key under which to store the computed loss.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Op in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
        average_loss: Whether to average the element-wise loss after the Loss Op.
    """
    def __init__(self,
                 inputs: Union[None, str, Iterable[str]] = None,
                 outputs: Union[None, str, Iterable[str]] = None,
                 mode: Union[None, str, Iterable[str]] = "!infer",
                 ds_id: Union[None, str, Iterable[str]] = None,
                 average_loss: bool = True):
        super().__init__(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id)
        self.average_loss = average_loss

    @property
    def true_key(self) -> str:
        return self.inputs[self.true_key_idx]

    @property
    def pred_key(self) -> str:
        return self.inputs[self.pred_key_idx]

    @property
    def true_key_idx(self) -> int:
        return 1

    @property
    def pred_key_idx(self) -> int:
        return 0



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import numpy as np

from fastestimator.op.numpyop.meta import Fuse
from fastestimator.op.numpyop.univariate import Minmax


class TestFuse(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.output_shape = (28, 28, 3)
        cls.multi_input = [np.random.randint(16, size=(28, 28, 3)), np.random.randint(16, size=(28, 28, 3))]

    def test_single_input(self):
        minmax = Minmax(inputs='x', outputs='y', mode='test')
        minmax2 = Minmax(inputs=["y", "z"], outputs="w", mode='test')
        fuse = Fuse([minmax, minmax2])
        with self.subTest('Check op inputs'):
            self.assertListEqual(fuse.inputs, ['x', 'z'])
        with self.subTest('Check op outputs'):
            self.assertListEqual(fuse.outputs, ['y', 'w'])
        with self.subTest('Check op mode'):
            self.assertSetEqual(fuse.mode, {'test'})
        output = fuse.forward(data=self.multi_input, state={"mode": "test"})
        with self.subTest('Check output type'):
            self.assertEqual(type(output), list)
        with self.subTest('Check output image shape'):
            self.assertEqual(output[0].shape, self.output_shape)



# Copyright 2020 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import unittest

import tensorflow as tf

from fastestimator.op.tensorop import LambdaOp
from fastestimator.op.tensorop.meta import Fuse


class TestFuse(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.output_shape = (28, 28, 3)
        cls.multi_input_tf = [tf.random.uniform(shape=(28, 28, 3)), tf.random.uniform(shape=(28, 28, 3))]

    def test_single_input_tf(self):
        a = LambdaOp(inputs='x', outputs='y', fn=lambda x: x + 1, mode='test')
        b = LambdaOp(inputs=['y', 'z'], outputs='w', fn=lambda x, y: x + y, mode='test')
        fuse = Fuse([a, b])
        with self.subTest('Check op inputs'):
            self.assertListEqual(fuse.inputs, ['x', 'z'])
        with self.subTest('Check op outputs'):
            self.assertListEqual(fuse.outputs, ['y', 'w'])
        with self.subTest('Check op mode'):
            self.assertSetEqual(fuse.mode, {'test'})
        output = fuse.forward(data=self.multi_input_tf, state={"mode": "test", "deferred": {}})
        with self.subTest('Check output type'):
            self.assertEqual(type(output), list)
        with self.subTest('Check output image shape'):
            self.assertEqual(output[0].shape, self.output_shape)

    @tf.function
    def test_single_input_tf_static(self):
        a = LambdaOp(inputs='x', outputs='y', fn=lambda x: x + 1, mode='test')
        b = LambdaOp(inputs=['y', 'z'], outputs='w', fn=lambda x, y: x + y, mode='test')
        fuse = Fuse([a, b])
        with self.subTest('Check op inputs'):
            self.assertListEqual(fuse.inputs, ['x', 'z'])
        with self.subTest('Check op outputs'):
            self.assertListEqual(fuse.outputs, ['y', 'w'])
        with self.subTest('Check op mode'):
            self.assertSetEqual(fuse.mode, {'test'})
        output = fuse.forward(data=self.multi_input_tf, state={"mode": "test", "deferred": {}})
        with self.subTest('Check output type'):
            self.assertEqual(type(output), list)
        with self.subTest('Check output image shape'):
            self.assertEqual(output[0].shape, self.output_shape)



# Copyright 2019 The FastEstimator Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import TypeVar

import numpy as np
import tensorflow as tf
import torch

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor, np.ndarray)


def zscore(data: Tensor, epsilon: float = 1e-7) -> Tensor:
    """Apply Zscore processing to a given tensor or array.

    This method can be used with Numpy data:
    ```python
    n = np.array([[0,1],[2,3]])
    b = fe.backend.zscore(n)  # [[-1.34164079, -0.4472136 ],[0.4472136 , 1.34164079]]
    ```

    This method can be used with TensorFlow tensors:
    ```python
    t = tf.constant([[0,1],[2,3]])
    b = fe.backend.zscore(t)  # [[-1.34164079, -0.4472136 ],[0.4472136 , 1.34164079]]
    ```

    This method can be used with PyTorch tensors:
    ```python
    p = torch.tensor([[0,1],[2,3]])
    b = fe.backend.zscore(p)  # [[-1.34164079, -0.4472136 ],[0.4472136 , 1.34164079]]
    ```

    Args:
        data: The input tensor or array.

    Returns:
        Data after substracting mean and divided by standard deviation.

    Raises:
        ValueError: If `tensor` is an unacceptable data type.
    """
    if tf.is_tensor(data):
        data = tf.cast(data, tf.float32)
        mean = tf.reduce_mean(data)
        std = tf.keras.backend.std(data)
        return (data - mean) / tf.maximum(std, epsilon)
    elif isinstance(data, torch.Tensor):
        data = data.type(torch.float32)
        mean = torch.mean(data)
        std = torch.std(data, unbiased=False)
        return (data - mean) / torch.max(std, torch.tensor(epsilon))
    elif isinstance(data, np.ndarray):
        mean = np.mean(data)
        std = np.std(data)
        return (data - mean) / max(std, epsilon)
    else:
        raise ValueError("Unrecognized data type {}".format(type(data)))



#  Copyright 2021 The FastEstimator Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ==============================================================================
from typing import Any, Dict, Iterable, Optional, TypeVar, Union

import cv2
import numpy as np
import tensorflow as tf
import torch

from fastestimator.backend.argmax import argmax
from fastestimator.backend.concat import concat
from fastestimator.backend.get_image_dims import get_image_dims
from fastestimator.backend.reduce_max import reduce_max
from fastestimator.backend.squeeze import squeeze
from fastestimator.trace.trace import Trace
from fastestimator.util.data import Data
from fastestimator.util.img_data import ImgData
from fastestimator.util.traceability_util import traceable
from fastestimator.util.util import to_number

Tensor = TypeVar('Tensor', tf.Tensor, torch.Tensor, np.ndarray)


@traceable()
class GradCAM(Trace):
    """A trace which draws GradCAM heatmaps on top of images.

    These are useful for visualizing supports for a model's classification. See https://arxiv.org/pdf/1610.02391.pdf
    for more details.

    Args:
        images: The key corresponding to images onto which to draw the CAM outputs.
        grads: The key corresponding to gradients of the model output with respect to a convolution layer of the model.
            You can easily extract these from any model by using the 'intermediate_layers' variable in a ModelOp, along
            with the GradientOp. Make sure to select a particular component of y_pred when computing gradients rather
            than using the entire vector. See our GradCAM XAI tutorial for an example.
        n_components: How many principal components to visualize.
        n_samples: How many images in total to display every epoch (or None to display all available images).
        labels: The key corresponding to the true labels of the images to be visualized.
        preds: The key corresponding to the model prediction for each image.
        label_mapping: {class_string: model_output_value}.
        outputs: The key into which to write the eigencam images.
        mode: What mode(s) to execute this Op in. For example, "train", "eval", "test", or "infer". To execute
            regardless of mode, pass None. To execute in all modes except for a particular one, you can pass an argument
            like "!infer" or "!train".
        ds_id: What dataset id(s) to execute this Trace in. To execute regardless of ds_id, pass None. To execute in all
            ds_ids except for a particular one, you can pass an argument like "!ds1".
    """
    def __init__(self,
                 images: str,
                 grads: str,
                 n_components: int = 3,
                 n_samples: Optional[int] = 5,
                 labels: Optional[str] = None,
                 preds: Optional[str] = None,
                 label_mapping: Optional[Dict[str, Any]] = None,
                 outputs: str = "gradcam",
                 mode: Union[None, str, Iterable[str]] = "!train",
                 ds_id: Union[None, str, Iterable[str]] = None):
        self.image_key = images
        self.grad_key = grads
        self.true_label_key = labels
        self.pred_label_key = preds
        inputs = [x for x in (images, grads, labels, preds) if x is not None]
        self.n_components = n_components
        self.n_samples = n_samples
        # TODO - handle non-hashable labels
        self.label_mapping = {val: key for key, val in label_mapping.items()} if label_mapping else None
        super().__init__(inputs=inputs, outputs=outputs, mode=mode, ds_id=ds_id)
        self.images = []
        self.grads = []
        self.labels = []
        self.preds = []
        self.n_found = 0

    def _reset(self) -> None:
        """Clear memory for next epoch.
        """
        self.images = []
        self.grads = []
        self.labels = []
        self.preds = []
        self.n_found = 0

    def on_batch_end(self, data: Data) -> None:
        if self.n_samples is None or self.n_found < self.n_samples:
            self.images.append(data[self.image_key])
            self.grads.append(data[self.grad_key])
            if self.true_label_key:
                self.labels.append(data[self.true_label_key])
            if self.pred_label_key:
                self.preds.append(data[self.pred_label_key])
            self.n_found += len(data[self.image_key])

    def on_epoch_end(self, data: Data) -> None:
        # Keep only the user-specified number of samples
        images = concat(self.images)[:self.n_samples or self.n_found]
        _, height, width = get_image_dims(images)
        grads = to_number(concat(self.grads)[:self.n_samples or self.n_found])
        if tf.is_tensor(images):
            grads = np.moveaxis(grads, source=-1, destination=1)  # grads should be channel first
        args = {}
        labels = None if not self.labels else concat(self.labels)[:self.n_samples or self.n_found]
        if labels is not None:
            if len(labels.shape) > 1:
                labels = argmax(labels, axis=-1)
            if self.label_mapping:
                labels = np.array([self.label_mapping[clazz] for clazz in to_number(squeeze(labels))])
            args[self.true_label_key] = labels
        preds = None if not self.preds else concat(self.preds)[:self.n_samples or self.n_found]
        if preds is not None:
            if len(preds.shape) > 1:
                preds = argmax(preds, axis=-1)
            if self.label_mapping:
                preds = np.array([self.label_mapping[clazz] for clazz in to_number(squeeze(preds))])
            args[self.pred_label_key] = preds
        args[self.image_key] = images
        # Clear memory
        self._reset()
        # Make the image
        # TODO: In future maybe allow multiple different grads to have side-by-side comparisons of classes
        components = [np.mean(grads, axis=1)]
        components = [np.maximum(component, 0) for component in components]
        masks = []
        for component_batch in components:
            img_batch = []
            for img in component_batch:
                img = cv2.resize(img, (height, width))
                img = img - np.min(img)
                img = img / np.max(img)
                img = cv2.cvtColor(cv2.applyColorMap(np.uint8(255 * img), cv2.COLORMAP_JET), cv2.COLOR_BGR2RGB)
                img = np.float32(img) / 255
                img_batch.append(img)
            img_batch = np.array(img_batch, dtype=np.float32)
            # Switch to channel first for pytorch
            if isinstance(images, torch.Tensor):
                img_batch = np.moveaxis(img_batch, source=-1, destination=1)
            masks.append(img_batch)

        components = [images + mask for mask in masks]  # This seems to work even if the image is 1 channel instead of 3
        components = [image / reduce_max(image) for image in components]

        for elem in components:
            args[self.grad_key] = elem

        result = ImgData(**args)
        data.write_without_log(self.outputs[0], result)
